import sys
_PY2 = sys.version_info[0] == 2

if _PY2:
    basestring = basestring
else:
    basestring = str


def init_config(config):
    if isinstance(config, basestring):
        with open(config, 'r') as f:
            content = f.read()
        import json
        config_dict = json.loads(content)
    elif isinstance(config, dict):
        config_dict = config
    else:
        config_dict = config.__dict__
    globals().update(config_dict)



import struct
import socket
import calendar
import pytz
import time
from datetime import datetime
import dateutil.parser

def ip_to_int(addr):
    return struct.unpack("!I", socket.inet_aton(addr))[0]

ip2int = ip_to_int    #Deprecated!

def int_to_ip(addr):
    return socket.inet_ntoa(struct.pack("!I", addr))

int2ip = int_to_ip    #Deprecated!

def timestamp_to_string(timestamp):
    return datetime.fromtimestamp(int(timestamp)).strftime('%Y-%m-%d %H:%M:%S')

time2string = timestamp_to_string    #Deprecated!

def string_to_timestamp(dt_string):
    return datetime_to_timestamp(dateutil.parser.parse(dt_string))

def get_tz(country_code):
    try:
        return pytz.timezone(pytz.country_timezones(country_code)[0])
    except:
        return pytz.timezone('Asia/Singapore')

def datetime_to_timestamp(dt):
    return calendar.timegm(dt.utctimetuple())

datetime_to_int = datetime_to_timestamp    #Deprecated!

def timestamp_to_datetime(timestamp, country_code):
    return datetime.fromtimestamp(timestamp, get_tz(country_code))

datetime_from_timestamp = timestamp_to_datetime

def get_current_datetime(country_code):
    now = int(time.time())
    return timestamp_to_datetime(now, country_code)

def strptime_with_tz(dt_string, dt_format, country_code):
    dt = datetime.strptime(dt_string, dt_format)
    return get_tz(country_code).localize(dt)



import sys
import random
import struct
import hashlib
import hmac

import platform
from Crypto.Cipher import AES

_PY2 = sys.version_info[0] == 2


if not _PY2:
    long = int
    xrange = range

# if platform.system().lower() == 'darwin':
#     from sakura.cpplib_osx import xtea_encrypt, xtea_decrypt, garena_xtea_encrypt, garena_xtea_decrypt
# else:
#     from sakura.cpplib import xtea_encrypt, xtea_decrypt, garena_xtea_encrypt, garena_xtea_decrypt

DWORD_BYTES = 4
QDWORD_BYTES = 8
UINT64_MASK = 0xffffffffffffffff


def random_byte():
    return chr(random.randint(0, 255))


def random_bytes(length):
    return ''.join([chr(random.randint(0, 255)) for _ in xrange(length)])

RANDOM_CHARACTER_SET = '1234567890abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ'


def random_string(length, allowed_chars=RANDOM_CHARACTER_SET):
    max_index = len(allowed_chars) - 1
    return ''.join([allowed_chars[random.randint(0, max_index)] for _ in xrange(length)])


def md5(plain_text):
    return hashlib.md5(plain_text).hexdigest()


def sha256(data):
    return hashlib.sha256(data).hexdigest()


def hmac_sha256(plain_text, key):
    return hmac.new(key, plain_text, hashlib.sha256).hexdigest()

"""
Reference:
https://en.wikipedia.org/wiki/Skipjack_(cipher)
https://github.com/boivie/skip32-java
"""
_SKIP32_FTABLE = [
    0xa3, 0xd7, 0x09, 0x83, 0xf8, 0x48,
    0xf6, 0xf4, 0xb3, 0x21, 0x15, 0x78, 0x99, 0xb1, 0xaf, 0xf9, 0xe7,
    0x2d, 0x4d, 0x8a, 0xce, 0x4c, 0xca, 0x2e, 0x52, 0x95, 0xd9, 0x1e,
    0x4e, 0x38, 0x44, 0x28, 0x0a, 0xdf, 0x02, 0xa0, 0x17, 0xf1, 0x60,
    0x68, 0x12, 0xb7, 0x7a, 0xc3, 0xe9, 0xfa, 0x3d, 0x53, 0x96, 0x84,
    0x6b, 0xba, 0xf2, 0x63, 0x9a, 0x19, 0x7c, 0xae, 0xe5, 0xf5, 0xf7,
    0x16, 0x6a, 0xa2, 0x39, 0xb6, 0x7b, 0x0f, 0xc1, 0x93, 0x81, 0x1b,
    0xee, 0xb4, 0x1a, 0xea, 0xd0, 0x91, 0x2f, 0xb8, 0x55, 0xb9, 0xda,
    0x85, 0x3f, 0x41, 0xbf, 0xe0, 0x5a, 0x58, 0x80, 0x5f, 0x66, 0x0b,
    0xd8, 0x90, 0x35, 0xd5, 0xc0, 0xa7, 0x33, 0x06, 0x65, 0x69, 0x45,
    0x00, 0x94, 0x56, 0x6d, 0x98, 0x9b, 0x76, 0x97, 0xfc, 0xb2, 0xc2,
    0xb0, 0xfe, 0xdb, 0x20, 0xe1, 0xeb, 0xd6, 0xe4, 0xdd, 0x47, 0x4a,
    0x1d, 0x42, 0xed, 0x9e, 0x6e, 0x49, 0x3c, 0xcd, 0x43, 0x27, 0xd2,
    0x07, 0xd4, 0xde, 0xc7, 0x67, 0x18, 0x89, 0xcb, 0x30, 0x1f, 0x8d,
    0xc6, 0x8f, 0xaa, 0xc8, 0x74, 0xdc, 0xc9, 0x5d, 0x5c, 0x31, 0xa4,
    0x70, 0x88, 0x61, 0x2c, 0x9f, 0x0d, 0x2b, 0x87, 0x50, 0x82, 0x54,
    0x64, 0x26, 0x7d, 0x03, 0x40, 0x34, 0x4b, 0x1c, 0x73, 0xd1, 0xc4,
    0xfd, 0x3b, 0xcc, 0xfb, 0x7f, 0xab, 0xe6, 0x3e, 0x5b, 0xa5, 0xad,
    0x04, 0x23, 0x9c, 0x14, 0x51, 0x22, 0xf0, 0x29, 0x79, 0x71, 0x7e,
    0xff, 0x8c, 0x0e, 0xe2, 0x0c, 0xef, 0xbc, 0x72, 0x75, 0x6f, 0x37,
    0xa1, 0xec, 0xd3, 0x8e, 0x62, 0x8b, 0x86, 0x10, 0xe8, 0x08, 0x77,
    0x11, 0xbe, 0x92, 0x4f, 0x24, 0xc5, 0x32, 0x36, 0x9d, 0xcf, 0xf3,
    0xa6, 0xbb, 0xac, 0x5e, 0x6c, 0xa9, 0x13, 0x57, 0x25, 0xb5, 0xe3,
    0xbd, 0xa8, 0x3a, 0x01, 0x05, 0x59, 0x2a, 0x46
]


def _skip32_g(key, k, w):
    g1 = w >> 8
    g2 = w & 0xff
    g3 = _SKIP32_FTABLE[g2 ^ (ord(key[(4 * k) % 10]) & 0xFF)] ^ g1
    g4 = _SKIP32_FTABLE[g3 ^ (ord(key[(4 * k + 1) % 10]) & 0xFF)] ^ g2
    g5 = _SKIP32_FTABLE[g4 ^ (ord(key[(4 * k + 2) % 10]) & 0xFF)] ^ g3
    g6 = _SKIP32_FTABLE[g5 ^ (ord(key[(4 * k + 3) % 10]) & 0xFF)] ^ g4

    return (g5 << 8) + g6


def skip32(key, value, encrypt):
    buf = [0]*4
    buf[0] = ((value >> 24) & 0xff)
    buf[1] = ((value >> 16) & 0xff)
    buf[2] = ((value >> 8) & 0xff)
    buf[3] = ((value >> 0) & 0xff)

    if encrypt:
        kstep = 1
        k = 0
    else:
        kstep = -1
        k = 23

    wl = (buf[0] << 8) + buf[1]
    wr = (buf[2] << 8) + buf[3]

    for i in range(12):
        wr ^= _skip32_g(key, k, wl) ^ k
        k += kstep
        wl ^= _skip32_g(key, k, wr) ^ k
        k += kstep

    buf[0] = (wr >> 8)
    buf[1] = (wr & 0xFF)
    buf[2] = (wl >> 8)
    buf[3] = (wl & 0xFF)

    return ((buf[0]) << 24) | ((buf[1]) << 16) | ((buf[2]) << 8) | (buf[3])


def skip32_encrypt(value, key):
    return skip32(key, value, True)


def skip32_decrypt(value, key):
    return skip32(key, value, False)


def xtea_encrypt_native(block, key, n=32):
    MASK = long(0xffffffff)
    DELTA = long(0x9e3779b9)
    v0 = block & MASK
    v1 = (block>>32) & MASK
    s = 0
    for _ in xrange(n):
        v0 = (v0 + ((((v1<<4) ^ (v1>>5)) + v1) ^ (s + key[s & 3]))) & MASK
        s = (s + DELTA) & MASK
        v1 = (v1 + ((((v0<<4) ^ (v0>>5)) + v0) ^ (s + key[(s>>11) & 3]))) & MASK
    return (v1<<32) | v0


def xtea_decrypt_native(block, key, n=32):
    MASK = long(0xffffffff)
    DELTA = long(0x9e3779b9)
    v0 = block & MASK
    v1 = (block >> 32) & MASK
    s = (DELTA * n) & MASK
    for _ in xrange(n):
        v1 = (v1 - ((((v0 << 4) ^ (v0 >> 5)) + v0) ^ (s + key[(s >> 11) & 3]))) & MASK
        s = (s - DELTA) & MASK
        v0 = (v0 - ((((v1 << 4) ^ (v1 >> 5)) + v1) ^ (s + key[s & 3]))) & MASK
    return (v1 << 32) | v0

XTEA_BLOCK_SIZE = 8


def xtea_cbc_encrypt(data, key, endian='<'):
    block_fmt = endian + 'Q'
    size = len(data)
    if size % XTEA_BLOCK_SIZE != 0:
        return None
    cipher = []
    block_cipher = 0
    for i in xrange(0, size, XTEA_BLOCK_SIZE):
        block = struct.unpack_from(block_fmt, data, i)[0]
        block_cipher = xtea_encrypt_native(block_cipher ^ block, key)
        cipher.append(struct.pack(block_fmt, block_cipher))
    return ''.join(cipher)


def xtea_cbc_decrypt(data, key, endian='<'):
    block_fmt = endian + 'Q'
    size = len(data)
    if size % XTEA_BLOCK_SIZE != 0:
        return None
    plain = []
    last_block = 0
    for i in xrange(0, size, XTEA_BLOCK_SIZE):
        block = struct.unpack_from(block_fmt, data, i)[0]
        plain_block = xtea_decrypt_native(block, key) ^ last_block
        plain.append(struct.pack(block_fmt, plain_block))
        last_block = block
    return ''.join(plain)

AES_BLOCK_SIZE = 16
AES_CBC_IV = '\0' * AES_BLOCK_SIZE


def aes_cbc_encrypt(data, key):
    return AES.new(key, AES.MODE_CBC, AES_CBC_IV).encrypt(data)


def aes_cbc_decrypt(data, key):
    try:
        return AES.new(key, AES.MODE_CBC, AES_CBC_IV).decrypt(data)
    except:
        return None


def pkcs7_padding(data, block_size):
    pad = block_size - (len(data) % block_size)
    data += chr(pad) * pad
    return data


def pkcs7_unpadding(data, block_size):
    size = len(data)
    if size < block_size or size % block_size != 0:
        return None
    pad = ord(data[-1])
    if pad <= 0 or pad > block_size:
        return None
    for i in xrange(2, pad+1):
        if ord(data[-i]) != pad:
            return None
    return data[:-pad]


GARENA_PADDING_MIN_BLOCK = 3


def garena_xtea_padding(data):
    data = random_bytes(XTEA_BLOCK_SIZE) + data
    data = pkcs7_padding(data, XTEA_BLOCK_SIZE)
    check_sum = 0
    for i in xrange(0, len(data), XTEA_BLOCK_SIZE):
        qword = struct.unpack_from('<Q', data, i)[0]
        check_sum += qword
    data += struct.pack('<Q', check_sum & UINT64_MASK)
    return data


def garena_xtea_unpadding(data):
    data_size = len(data)
    if data_size % XTEA_BLOCK_SIZE != 0:
        return None
    if data_size < XTEA_BLOCK_SIZE * GARENA_PADDING_MIN_BLOCK:
        return None
    check_sum = struct.unpack_from('<Q', data, len(data) - XTEA_BLOCK_SIZE)[0]
    for i in xrange(0, len(data) - XTEA_BLOCK_SIZE, XTEA_BLOCK_SIZE):
        qword = struct.unpack_from('<Q', data, i)[0]
        check_sum -= qword
    if (check_sum & UINT64_MASK) != 0:
        return None
    return pkcs7_unpadding(data[XTEA_BLOCK_SIZE:-XTEA_BLOCK_SIZE], XTEA_BLOCK_SIZE)


def garena_xtea_encrypt_native(data, key):
    padding_data = garena_xtea_padding(data)
    return xtea_cbc_encrypt(padding_data, key)


def garena_xtea_decrypt_native(data, key):
    data_size = len(data)
    if data_size % XTEA_BLOCK_SIZE != 0:
        return None
    if data_size < XTEA_BLOCK_SIZE * GARENA_PADDING_MIN_BLOCK:
        return None
    padding_data = xtea_cbc_decrypt(data, key)
    return garena_xtea_unpadding(padding_data)


def garena_cbc_padding(data, block_size):
    data = random_bytes(block_size) + data
    data = pkcs7_padding(data, block_size)
    check_sum = [0] * block_size
    i = 0
    for ch in data:
        check_sum[i] ^= ord(ch)
        i += 1
        if i >= block_size:
            i = 0
    data = ''.join([data] + [chr(v) for v in check_sum])
    return data


def garena_cbc_unpadding(data, block_size):
    data_size = len(data)
    if data_size % block_size != 0:
        return None
    if data_size < block_size * GARENA_PADDING_MIN_BLOCK:
        return None
    check_sum = [0] * block_size
    i = 0
    for ch in data:
        check_sum[i] ^= ord(ch)
        i += 1
        if i >= block_size:
            i = 0
    for v in check_sum:
        if v != 0:
            return None
    return pkcs7_unpadding(data[block_size:-block_size], block_size)


def garena_aes_encrypt(data, key):
    padding_data = garena_cbc_padding(data, AES_BLOCK_SIZE)
    return aes_cbc_encrypt(padding_data, key)


def garena_aes_decrypt(data, key):
    data_size = len(data)
    if data_size % AES_BLOCK_SIZE != 0:
        return None
    if data_size < AES_BLOCK_SIZE * GARENA_PADDING_MIN_BLOCK:
        return None
    padding_data = aes_cbc_decrypt(data, key)
    return garena_cbc_unpadding(padding_data, AES_BLOCK_SIZE)


def create_django_secret_key():
    return random_string(50, 'abcdefghijklmnopqrstuvwxyz0123456789!@#$%^&*(-_=+)')



import struct

class BufferWriter():

    def __init__(self,  endian='<'):
        self._buffer = []
        self._endian = endian

    @property
    def buffer(self):
        return b''.join(self._buffer)

    def set_endian(self, endian):
        self._endian = endian

    def clear(self):
        self._buffer = []

    def add_int8(self, data):
        self._add_int('b', data)

    def add_uint8(self, data):
        self._add_int('B', data)

    def add_int16(self, data):
        self._add_int('h', data)

    def add_uint16(self, data):
        self._add_int('H', data)

    def add_int32(self, data):
        self._add_int('i', data)

    def add_uint32(self, data):
        self._add_int('I', data)

    def add_int64(self, data):
        self._add_int('q', data)

    def add_uint64(self, data):
        self._add_int('Q', data)

    def add_float(self, data):
        self._add_int('f', data)

    def add_double(self, data):
        self._add_int('d', data)

    def add_buffer(self, buf):
        self._buffer.append(buf)

    def add_padding(self, size):
        self._buffer.append('\0' * size)

    def _add_int(self, fmt, data):
        self._buffer.append(struct.pack(self._endian + fmt, data))



import sys
_PY3 = sys.version_info[0] == 3

if _PY3:
    long = int
    unicode = str
    xrange = range
    integer_types = (int,)
    basestring = str
    import binascii
    byteshex = lambda b: binascii.hexlify(b)
    bytesunhex = lambda b: binascii.unhexlify(b)

    def v_iteritems(d):
        return d.items()

else:

    byteshex = lambda b: b.encode("hex")
    bytesunhex = lambda b: b.decode("hex")
    xrange = xrange
    integer_types = (int, long)

    def v_iteritems(d):
        return d.iteritems()



import socket
import struct
import json
import platform
from sakura.logger import log
from sakura.version_tools import byteshex


class GTcpClient(object):

    def __init__(self, address, port, timeout=10, retry=True):
        self._address = address
        self._port = port
        self._socket = None
        self._timeout = timeout
        self._retry = retry

    def close(self):
        if self._socket is not None:
            self._socket.close()
            self._socket = None

    def send(self, request):
        if self._socket is None:
            if not self._connect():
                return False
        packet = struct.pack('<I%ds' % len(request), len(request), request)
        try:
            self._socket.sendall(packet)
            return True
        except Exception as ex:
            self.close()
            if not self._retry:
                # log.exception('tcp_send_fail|error=send_fail,address=%s,port=%u,request=%s', self._address, self._port, request.encode('hex'))
                log.exception('tcp_send_fail|error=send_fail,address=%s,port=%u,request=%s', self._address, self._port, byteshex(request))
                return False
            # log.warn('tcp_send_fail|error=send_fail_will_retry,address=%s,port=%u,retry=0,request=%s,ex=%s', self._address, self._port, request.encode('hex'), ex)
            log.warn('tcp_send_fail|error=send_fail_will_retry,address=%s,port=%u,retry=0,request=%s,ex=%s',
                     self._address, self._port, byteshex(request), ex)
            if not self._connect():
                # log.exception('tcp_send_fail|error=retry_reconnect,address=%s,port=%u,retry=0,request=%s', self._address, self._port, request.encode('hex'))
                log.exception('tcp_send_fail|error=retry_reconnect,address=%s,port=%u,retry=0,request=%s',
                              self._address, self._port, byteshex(request))
                return False
            try:
                self._socket.sendall(packet)
                return True
            except Exception as ex:
                # log.exception('tcp_send_fail|error=retry_send_fail,address=%s,port=%u,retry=1,request=%s', self._address, self._port, request.encode('hex'))
                log.exception('tcp_send_fail|error=retry_send_fail,address=%s,port=%u,retry=1,request=%s',
                              self._address, self._port, byteshex(request))
                self.close()
                return False

    def receive(self):
        if self._socket is None:
            if not self._connect():
                return None
        try:
            length_data = self._recv(4)
            length = struct.unpack('<I', length_data)[0]
            return self._recv(length)
        except Exception as ex:
            log.warn('tcp_recv_fail|address=%s,port=%u,ex=%s', self._address, self._port, ex, exc_info=True)
            self.close()
            return None

    def request(self, request):
        if self._socket is None:
            if not self._connect():
                return None
        packet = struct.pack('<I%ds' % len(request), len(request), request)
        try:
            self._socket.sendall(packet)
            length_data = self._recv(4)
        except Exception as ex:
            self.close()
            if not self._retry:
                # log.exception('tcp_request_fail|error=recv_length_fail,address=%s,port=%u,request=%s', self._address, self._port, request.encode('hex'))
                log.exception('tcp_request_fail|error=recv_length_fail,address=%s,port=%u,request=%s', self._address,
                              self._port, byteshex(request))
                return None
            if isinstance(ex, socket.timeout):
                # log.exception('tcp_request_fail|error=recv_length_timeout,address=%s,port=%u,retry=0,request=%s', self._address, self._port, request.encode('hex'))
                log.exception('tcp_request_fail|error=recv_length_timeout,address=%s,port=%u,retry=0,request=%s',
                              self._address, self._port, byteshex(request))
                return None
            # log.warn('tcp_request_fail|error=recv_length_fail_will_retry,address=%s,port=%u,retry=0,request=%s,ex=%s', self._address, self._port, request.encode('hex'), ex)
            log.warn('tcp_request_fail|error=recv_length_fail_will_retry,address=%s,port=%u,retry=0,request=%s,ex=%s',
                     self._address, self._port, byteshex(request), ex)
            if not self._connect():
                # log.exception('tcp_request_fail|error=retry_reconnect,address=%s,port=%u,retry=0,request=%s', self._address, self._port, request.encode('hex'))
                log.exception('tcp_request_fail|error=retry_reconnect,address=%s,port=%u,retry=0,request=%s',
                              self._address, self._port, byteshex(request))
                return None
            try:
                self._socket.sendall(packet)
                length_data = self._recv(4)
            except Exception as ex:
                # log.exception('tcp_request_fail|error=retry_recv_length_fail,address=%s,port=%u,retry=1,request=%s', self._address, self._port, request.encode('hex'))
                log.exception('tcp_request_fail|error=retry_recv_length_fail,address=%s,port=%u,retry=1,request=%s',
                              self._address, self._port, byteshex(request))
                self.close()
                return None
        try:
            length = struct.unpack('<I', length_data)[0]
            return self._recv(length)
        except Exception as ex:
            # log.exception('tcp_request_fail|error=recv_data_fail,address=%s,port=%u,request=%s', self._address, self._port, request.encode('hex'))
            log.exception('tcp_request_fail|error=recv_data_fail,address=%s,port=%u,request=%s', self._address,
                          self._port, byteshex(request))
            self.close()
            return None

    def request_json(self, request):
        request_data = json.dumps(request)
        reply = self.request(request_data)
        if reply is None:
            return None
        try:
            return json.loads(reply)
        except Exception as ex:
            log.warn('tcp_request_parse_json_fail|address=%s,port=%u,reply=%s,ex=%s', self._address, self._port, reply, ex)
            self.close()
            return None

    def _connect(self):
        try:
            self._socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            if self._timeout > 0:
                self._socket.settimeout(self._timeout)
            self._set_keep_alive()
            self._socket.connect((self._address, self._port))
            log.info('tcp_connect|address=%s,port=%u', self._address, self._port)
            return True
        except Exception as ex:
            log.exception('tcp_connect_fail|address=%s,port=%u,ex=%s', self._address, self._port, ex)
            self.close()
            return False

    def _recv(self, length):
        data = b''
        while length > 0:
            recv_data = self._socket.recv(length)
            recv_length = len(recv_data)
            if recv_length <= 0:
                raise Exception('socket_closed')
            data += recv_data
            length -= recv_length
        return data

    def _set_keep_alive(self):
        self._socket.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)
        if platform.system().lower() == 'linux':
            self._socket.setsockopt(socket.SOL_TCP, socket.TCP_KEEPIDLE, 30)
            self._socket.setsockopt(socket.SOL_TCP, socket.TCP_KEEPINTVL, 5)
            self._socket.setsockopt(socket.SOL_TCP, socket.TCP_KEEPCNT, 5)



from logging.handlers import *
from sakura import ipc
import os, sys, traceback
import platform

if platform.system().lower() == 'windows':

    class _IncrementalLock(ipc.Lock):

        def try_lock(self, value):
            self.acquire()
            return True

        def release_lock(self, value):
            self.release()

else:

    class _IncrementalLock():

        def __init__(self, name):
            self.name = name
            self.lock = None

        def try_lock(self, value):
            if self.lock is not None:
                return False
            self.lock = ipc.Lock(self.name)
            self.lock.acquire()
            old_value = self.lock.handle.read()
            if value <= old_value:
                self.lock.release()
                self.lock = None
                return False
            return True

        def release_lock(self, value):
            if self.lock is None:
                return
            with open(self.lock.filename, 'w') as f:
                f.write(value)
            self.lock.release()
            self.lock = None

class MPRotatingFileHandler(RotatingFileHandler, object):

    def __init__(self, filename, mode='a', maxBytes=0, backupCount=0, encoding=None, delay=0):
        self.mplock_name = filename.replace('/', '_').replace('\\', '_')
        super(MPRotatingFileHandler, self).__init__(filename, mode, maxBytes, backupCount, encoding, delay)

    def doRollover(self):
        mplock = ipc.Lock(self.mplock_name)
        mplock.acquire()
        if not self.prepareRollover():
            mplock.release()
            return
        if self.stream:
            self.stream.close()
            self.stream = None
        if self.backupCount > 0:
            for i in range(self.backupCount - 1, 0, -1):
                sfn = "%s.%d" % (self.baseFilename, i)
                dfn = "%s.%d" % (self.baseFilename, i + 1)
                if os.path.exists(sfn):
                    if os.path.exists(dfn):
                        os.remove(dfn)
                    os.rename(sfn, dfn)
            dfn = self.baseFilename + ".1"
            if os.path.exists(dfn):
                os.remove(dfn)
            os.rename(self.baseFilename, dfn)
        mplock.release()
        self.stream = self._open()

    def prepareRollover(self):
        if self.stream is not None:
            self.flush()
            self.stream.close()
            self.stream = None
        self.stream = self._open()
        if self.maxBytes > 0:
            self.stream.seek(0, 2)
            if self.stream.tell() >= float(self.maxBytes) * 0.99:
                return True
        return False

class MPTimedRotatingFileHandler(TimedRotatingFileHandler, object):

    def __init__(self, filename, when='h', interval=1, backupCount=0, encoding=None, delay=False, utc=False):
        self.mplock_name = filename.replace('/', '_').replace('\\', '_')
        super(MPTimedRotatingFileHandler, self).__init__(filename, when, interval, backupCount, encoding, delay, utc)

    def computeRollover(self, currentTime):
        result = int(super(MPTimedRotatingFileHandler, self).computeRollover(currentTime))
        if self.when == 'S':
            result = result
        elif self.when == 'M':
            result = result / 60 * 60
        elif self.when == 'H':
            result = result / 3600 * 3600
        return result

    def doRollover(self):
        if self.stream:
            self.stream.close()
            self.stream = None
        t = self.rolloverAt - self.interval
        if self.utc:
            timeTuple = time.gmtime(t)
        else:
            timeTuple = time.localtime(t)
        formattedTime = time.strftime(self.suffix, timeTuple)
        mplock = _IncrementalLock(self.mplock_name)
        if mplock.try_lock(formattedTime):
            dfn = self.baseFilename + "." + formattedTime
            if not os.path.exists(dfn):
                os.rename(self.baseFilename, dfn)
                if self.backupCount > 0:
                    for s in self.getFilesToDelete():
                        os.remove(s)
            mplock.release_lock(formattedTime)

        self.stream = self._open()
        currentTime = int(time.time())
        newRolloverAt = self.computeRollover(currentTime)
        while newRolloverAt <= currentTime:
            newRolloverAt = newRolloverAt + self.interval
        if (self.when == 'MIDNIGHT' or self.when.startswith('W')) and not self.utc:
            dstNow = time.localtime(currentTime)[-1]
            dstAtRollover = time.localtime(newRolloverAt)[-1]
            if dstNow != dstAtRollover:
                if not dstNow:  # DST kicks in before next rollover, so we need to deduct an hour
                    newRolloverAt -= 3600
                else:           # DST bows out before next rollover, so we need to add an hour
                    newRolloverAt += 3600
        self.rolloverAt = newRolloverAt



from tornado.tcpclient import TCPClient
import socket
import platform
import struct
from sakura.logger import log
from sakura import event_loop
from sakura.version_tools import byteshex


class GTcpAsyncClient(object):

    HEADER_SIZE = 4
    TCP_MAX_PACKET_SIZE = 256 * 1024

    def __init__(self, id, address, port, on_receive_packet, on_connect=None, on_disconnect=None):
        self._id = id
        self._address = address
        self._port = port
        self._on_receive_packet_callback = on_receive_packet
        self._on_connect_callback = on_connect
        self._on_disconnect_callback = on_disconnect
        self._stream = None
        self._pending_buffer = []
        self._async_connect()

    @property
    def id(self):
        return self._id

    def send(self, packet):
        packet_length = struct.pack('!I', len(packet))
        if self._stream is not None:
            self._stream.write(packet_length + packet)
        else:
            self._pending_buffer.append(packet_length)
            self._pending_buffer.append(packet)

    def close(self):
        event_loop.add_callback(self._close)

    def clear_buffer(self):
        self._pending_buffer = []

    def _async_connect(self):
        log.info('tcp_asnyc_client_try_connect|id=%s,address=%s,port=%s', self._id, self._address, self._port)
        TCPClient().connect(self._address, self._port).add_done_callback(self._on_connect)

    def _on_connect(self, future):
        try:
            self._stream = future.result()
        except:
            log.exception('tcp_asnyc_client_connect_fail|id=%s,address=%s,port=%s', self._id, self._address, self._port)
            self._close()
            self._async_connect()
            return
        log.info('tcp_asnyc_client_connect|id=%s,address=%s,port=%s', self._id, self._address, self._port)
        self._stream.set_close_callback(self._on_close)
        self._set_keep_alive()
        if self._on_connect_callback is not None:
            try:
                self._on_connect_callback(self)
            except:
                log.exception('tcp_asnyc_client_on_connect_exception|id=%s,address=%s,port=%s', self._id, self._address, self._port)
        if self._pending_buffer:
            self._stream.write(b''.join(self._pending_buffer))
            self._pending_buffer = []
        self._recv_header()

    def _close(self):
        if self._stream:
            self._stream.close()
            self._stream = None

    def _on_close(self):
        log.info('tcp_asnyc_client_disconnect|id=%s,address=%s,port=%s', self._id, self._address, self._port)
        if self._on_disconnect_callback is not None:
            try:
                self._on_disconnect_callback(self)
            except:
                log.exception('tcp_asnyc_client_on_disconnect_exception|id=%s,address=%s,port=%s', self._id, self._address, self._port)
        self._async_connect()

    def _recv_header(self):
        self._stream.read_bytes(self.HEADER_SIZE, self._on_recv_header)

    def _on_recv_header(self, data):
        (body_size,) = struct.unpack('!I', data)
        if body_size > self.TCP_MAX_PACKET_SIZE:
            log.error('tcp_asnyc_client_body_size_overflow|id=%s,size=%u', self._id, body_size)
            self._close()
            return
        self._stream.read_bytes(body_size, self._on_recv_body)

    def _on_recv_body(self, data):
        if self._on_receive_packet_callback is not None:
            try:
                self._on_receive_packet_callback(self, data)
            except:
                # log.exception('tcp_asnyc_client_on_receive_exception|id=%s,packet=%s', self._id, data.encode('hex'))
                log.exception('tcp_asnyc_client_on_receive_exception|id=%s,packet=%s', self._id, byteshex(data))
        self._recv_header()

    def _set_keep_alive(self):
        stream_socket = self._stream.socket
        stream_socket.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)
        if platform.system().lower() == 'linux':
            stream_socket.setsockopt(socket.SOL_TCP, socket.TCP_KEEPIDLE, 30)
            stream_socket.setsockopt(socket.SOL_TCP, socket.TCP_KEEPINTVL, 5)
            stream_socket.setsockopt(socket.SOL_TCP, socket.TCP_KEEPCNT, 5)



import sys
import re
try:
    import cPickle as pickle
except ImportError:  # pragma: no cover
    import pickle
import random
from collections import defaultdict
import platform
import re
from sakura.conhash import ConHash

_PY2 = sys.version_info[0] == 2


if _PY2:
    _integer_types = (int, long)
    _iteritems = lambda d: d.iteritems()

    def _to_native(x, charset='utf-8', errors='strict'):
        if x is None or isinstance(x, str):
            return x
        return x.encode(charset, errors)
else:
    _integer_types = (int, )
    xrange = range

    _iteritems = lambda d: iter(d.items())


    def _to_native(x, charset='utf-8', errors='strict'):
        if x is None or isinstance(x, str):
            return x
        return x.decode(charset, errors)

def _items(dictorseq):
    """Wrapper for efficient iteration over dict represented by dicts
    or sequences::

        >>> for k, v in _items((i, i*i) for i in xrange(5)):
        ...    assert k*k == v

        >>> for k, v in _items(dict((i, i*i) for i in xrange(5))):
        ...    assert k*k == v

    """
    if hasattr(dictorseq, 'items'):
        return _iteritems(dictorseq)
    return dictorseq

_DEFAULT_CACHE_TIMEOUT = 300
_DEFAULT_SOCKET_TIMEOUT = 3

if platform.system().lower() == 'linux':
    import socket
    _TCP_KEEP_ALIVE_OPTIONS = {
        socket.TCP_KEEPIDLE: 30,
        socket.TCP_KEEPINTVL: 5,
        socket.TCP_KEEPCNT: 5,
    }
else:
    _TCP_KEEP_ALIVE_OPTIONS = {}

class BaseCache(object):
    """Baseclass for the cache systems.  All the cache systems implement this
    API or a superset of it.

    :param default_timeout: the default timeout (in seconds) that is used if no
                            timeout is specified on :meth:`set`.
    """

    def __init__(self, config):
        self._client = None
        self.default_timeout = config.get('default_timeout', _DEFAULT_CACHE_TIMEOUT)

    @property
    def raw_client(self):
        """Get raw cache client.

        :returns: Underlying cache client object.
        """
        return self._client

    def get(self, key):
        """Look up key in the cache and return the value for it.

        :param key: the key to be looked up.
        :returns: The value if it exists and is readable, else ``None``.
        """
        return None

    def delete(self, key, noreply=False):
        """Delete `key` from the cache.

        :param key: the key to delete.
        :param noreply: instructs the server to not send the reply.
        :returns: Whether the key has been deleted.
        :rtype: boolean
        """
        return True

    def get_list(self, keys):
        """Returns a list of values for the given keys.
        For each key a item in the list is created::

            foo, bar = cache.get_list("foo", "bar")

        Has the same error handling as :meth:`get`.

        :param keys: The function accepts multiple keys as positional
                     arguments.
        """
        values = self.get_many(keys)
        return [values.get(key) for key in keys]

    def get_many(self, keys):
        """Like :meth:`get_list` but return a dict::
        If the given key is missing, it will be missing from the response dict.

            d = cache.get_many("foo", "bar")
            foo = d["foo"]
            bar = d["bar"]

        :param keys: The function accepts multiple keys as positional
                     arguments.
        """
        return dict([(key, self.get(key)) for key in keys])

    def set(self, key, value, timeout=None, noreply=False):
        """Add a new key/value to the cache (overwrites value, if key already
        exists in the cache).

        :param key: the key to set
        :param value: the value for the key
        :param timeout: the cache timeout for the key.
                        If not specified, it uses the default timeout.
                        If specified 0, it will never expire.
        :param noreply: instructs the server to not send the reply.
        :returns: Whether the key existed and has been set.
        :rtype: boolean
        """
        return True

    def add(self, key, value, timeout=None, noreply=False):
        """Works like :meth:`set` but does not overwrite the values of already
        existing keys.

        :param key: the key to set
        :param value: the value for the key
        :param timeout: the cache timeout for the key.
                        If not specified, it uses the default timeout.
                        If specified 0, it will never expire.
        :param noreply: instructs the server to not send the reply.
        :returns: Same as :meth:`set`, but also ``False`` for already
                  existing keys.
        :rtype: boolean
        """
        return True

    def set_many(self, data, timeout=None, noreply=False):
        """Sets multiple keys and values from a dict.

        :param data: a dict with the keys/values to set.
        :param timeout: the cache timeout for the key.
                        If not specified, it uses the default timeout.
                        If specified 0, it will never expire.
        :param noreply: instructs the server to not send the reply.
        :returns: Whether all given keys have been set.
        :rtype: boolean
        """
        rv = True
        for key, value in _items(data):
            if not self.set(key, value, timeout):
                rv = False
        return rv

    def delete_many(self, keys, noreply=False):
        """Deletes multiple keys at once.

        :param keys: The function accepts multiple keys as positional
                     arguments.
        :param noreply: instructs the server to not send the reply.
        :returns: Whether all given keys have been deleted.
        :rtype: boolean
        """
        return all(self.delete(key) for key in keys)

    def clear(self):
        """Clears the cache.  Keep in mind that not all caches support
        completely clearing the cache.
        :returns: Whether the cache has been cleared.
        :rtype: boolean
        """
        return True

    def incr(self, key, delta=1, noreply=False):
        """Increments the value of a key by `delta`.  If the key does
        not yet exist it is initialized with `delta`.

        For supporting caches this is an atomic operation.

        :param key: the key to increment.
        :param delta: the delta to add.
        :param noreply: instructs the server to not send the reply.
        :returns: The new value or ``None`` for backend errors.
        """
        value = (self.get(key) or 0) + delta
        return value if self.set(key, value) else None

    def decr(self, key, delta=1, noreply=False):
        """Decrements the value of a key by `delta`.  If the key does
        not yet exist it is initialized with `-delta`.

        For supporting caches this is an atomic operation.

        :param key: the key to increment.
        :param delta: the delta to subtract.
        :param noreply: instructs the server to not send the reply.
        :returns: The new value or `None` for backend errors.
        """
        value = (self.get(key) or 0) - delta
        return value if self.set(key, value) else None

    def hgetall(self, key):
        """Look up hash in the cache and return all fields and values for it.

        :param key: the key of hash to be looked up.
        :returns: The dict value of hash, if empty, return {}.
        :rtype: dict
        """
        raise NotImplementedError()

    def hget(self, key, field):
        """Look up field in the hash and return the value for it.

        :param key: the key of hash to be looked up.
        :param field: the filed in the hash to be looked up.
        :returns: The value if it exists and is readable, else ``None``.
        """
        raise NotImplementedError()

    def hset(self, key, field, value, timeout=None, noreply=False):
        """Add a new filed/value to the hash in cache (overwrites value, if key already
        exists in the cache).

        :param key: the key of hash to set
        :param key: the field in the hash to set
        :param value: the value for the field
        :param timeout: the cache timeout for the field.
                        If not specified, it uses the default timeout.
                        If specified 0, it will never expire.
        :param noreply: instructs the server to not send the reply.
        :returns: Whether the key existed and has been set.
        :rtype: boolean
        """
        raise NotImplementedError()

    def hdel(self, key, field, noreply=False):
        """Delete field of hash from the cache.

        :param key: the key of hash to delete
        :param key: the field in the hash to delete
        :param noreply: instructs the server to not send the reply.
        :returns: Whether the key has been deleted.
        :rtype: boolean
        """
        raise NotImplementedError()

class NullCache(BaseCache):
    """A cache that doesn't cache.  This can be useful for unit testing.

    :param default_timeout: a dummy parameter that is ignored but exists
                            for API compatibility with other caches.
    """
    def __init__(self, config):
        super(NullCache, self).__init__(config)


_test_memcached_key = re.compile(r'[^\x00-\x21\xff]{1,250}$').match
_PYLIBMC_BEHAVIORS = {
    'connect_timeout': _DEFAULT_SOCKET_TIMEOUT * 1000,
    'send_timeout': _DEFAULT_SOCKET_TIMEOUT * 1000 * 1000,
    'receive_timeout': _DEFAULT_SOCKET_TIMEOUT * 1000 * 1000,
}

class MemcachedCache(BaseCache):
    """A cache that uses memcached as backend.

    This cache looks into the following packages/modules to find bindings for
    memcached:

        - ``pylibmc``
        - ``google.appengine.api.memcached``
        - ``memcached``

    Implementation notes:  This cache backend works around some limitations in
    memcached to simplify the interface.  For example unicode keys are encoded
    to utf-8 on the fly.  Methods such as :meth:`~BaseCache.get_many` return
    the keys in the same format as passed.  Furthermore all get methods
    silently ignore key errors to not cause problems when untrusted user data
    is passed to the get methods which is often the case in web applications.

    :param host: memcached server host.
    :param port: memcached server port.
    :param default_timeout: the default timeout that is used if no timeout is
                            specified on :meth:`~BaseCache.set`.
    :param key_prefix: a prefix that is added before all keys.  This makes it
                       possible to use the same memcached server for different
                       applications.  Keep in mind that
                       :meth:`~BaseCache.clear` will also clear keys with a
                       different prefix.
    """

    def __init__(self, config):
        BaseCache.__init__(self, config)
        servers = ['%s:%d' % (config['host'], config['port'])]
        self.key_prefix = _to_native(config.get('key_prefix', None))
        if servers is None:
            servers = ['127.0.0.1:11211']
        self._client = self.import_preferred_memcache_lib(servers)
        if self._client is None:
            raise RuntimeError('no memcache module found')

    def _normalize_key(self, key):
        key = _to_native(key, 'utf-8')
        if self.key_prefix:
            key = self.key_prefix + key
        return key

    def _normalize_timeout(self, timeout):
        if timeout is None:
            timeout = self.default_timeout
        return timeout

    def get(self, key):
        key = self._normalize_key(key)
        # memcached doesn't support keys longer than that.  Because often
        # checks for so long keys can occur because it's tested from user
        # submitted data etc we fail silently for getting.
        if _test_memcached_key(key):
            return self._client.get(key)

    def get_many(self, keys):
        key_mapping = {}
        have_encoded_keys = False
        for key in keys:
            encoded_key = self._normalize_key(key)
            if not isinstance(key, str):
                have_encoded_keys = True
            if _test_memcached_key(key):
                key_mapping[encoded_key] = key
        d = rv = self._client.get_multi(key_mapping.keys())
        if have_encoded_keys or self.key_prefix:
            rv = {}
            for key, value in _iteritems(d):
                rv[key_mapping[key]] = value
        return rv

    def add(self, key, value, timeout=None, noreply=False):
        key = self._normalize_key(key)
        timeout = self._normalize_timeout(timeout)
        return self._client.add(key, value, timeout)

    def set(self, key, value, timeout=None, noreply=False):
        key = self._normalize_key(key)
        timeout = self._normalize_timeout(timeout)
        return self._client.set(key, value, timeout)

    def set_many(self, data, timeout=None, noreply=False):
        new_data = {}
        for key, value in _items(data):
            key = self._normalize_key(key)
            new_data[key] = value

        timeout = self._normalize_timeout(timeout)
        failed_keys = self._client.set_multi(new_data, timeout)
        return not failed_keys

    def delete(self, key, noreply=False):
        key = self._normalize_key(key)
        if _test_memcached_key(key):
            return self._client.delete(key) is not 0
        else:
            return False

    def delete_many(self, keys, noreply=False):
        new_keys = []
        for key in keys:
            key = self._normalize_key(key)
            if _test_memcached_key(key):
                new_keys.append(key)
        return self._client.delete_multi(new_keys) is not 0

    def clear(self):
        return self._client.flush_all()

    def incr(self, key, delta=1, noreply=False):
        key = self._normalize_key(key)
        return self._client.incr(key, delta)

    def decr(self, key, delta=1, noreply=False):
        key = self._normalize_key(key)
        return self._client.decr(key, delta)

    def import_preferred_memcache_lib(self, servers):
        """Returns an initialized memcache client.  Used by the constructor."""
        try:
            import pylibmc
        except ImportError:
            pass
        else:
            return pylibmc.Client(servers, behaviors=_PYLIBMC_BEHAVIORS)

        try:
            import memcache
        except ImportError:
            pass
        else:
            return memcache.Client(servers)


try:

    from pymemcache.client import Client as PyMemcachedClient
    from pymemcache.serde import python_memcache_serializer, python_memcache_deserializer

    class PyMemcachedCache(BaseCache):
        """A cache client based on pymemcache. implemented by pure python and support noreply.
        """
        def __init__(self, config):
            BaseCache.__init__(self, config)
            self._client = PyMemcachedClient((config['host'], config['port']),
                serializer=python_memcache_serializer, deserializer=python_memcache_deserializer,
                connect_timeout=_DEFAULT_SOCKET_TIMEOUT, timeout=_DEFAULT_SOCKET_TIMEOUT,
                key_prefix=config.get('key_prefix', ''))

        def get(self, key):
            return self._client.get(key)

        def delete(self, key, noreply=False):
            self._client.delete(key, noreply)
            return True

        def get_many(self, keys):
            return self._client.get_many(keys)

        def set(self, key, value, timeout=None, noreply=False):
            if timeout is None:
                timeout = self.default_timeout
            return self._client.set(key, value, timeout, noreply)

        def add(self, key, value, timeout=None, noreply=False):
            if timeout is None:
                timeout = self.default_timeout
            return self._client.add(key, value, timeout, noreply)

        def set_many(self, data, timeout=None, noreply=False):
            if timeout is None:
                timeout = self.default_timeout
            return self._client.set_many(data, timeout, noreply)

        def delete_many(self, keys, noreply=False):
            return self._client.delete_many(keys, noreply)

        def clear(self):
            return self._client.flush_all(noreply=False)

        def incr(self, key, delta=1, noreply=False):
            return self._client.incr(key, delta, noreply)

        def decr(self, key, delta=1, noreply=False):
            return self._client.decr(key, delta, noreply)

except ImportError:

    PyMemcachedCache = MemcachedCache

_redis_kwargs_exclusions = set(('id', 'type', 'replica', 'default', 'default_timeout', 'key_prefix'))
_DEFAULT_REDIS_BLOCKING_POOL_TIMEOUT = 5

class RedisCache(BaseCache):
    """Uses the Redis key-value store as a cache backend.

    :param host: address of the Redis server or an object which API is
                 compatible with the official Python Redis client (redis-py).
    :param port: port number on which Redis server listens for connections.
    :param password: password authentication for the Redis server.
    :param db: db (zero-based numeric index) on Redis Server to connect.
    :param default_timeout: the default timeout that is used if no timeout is
                            specified on :meth:`~BaseCache.set`.
    :param key_prefix: A prefix that should be added to all keys.

    Any additional keyword arguments will be passed to ``redis.Redis``.
    """

    def __init__(self, config):
        BaseCache.__init__(self, config)
        self.key_prefix = config.get('key_prefix', '')
        try:
            import redis
        except ImportError:
            raise RuntimeError('no redis module found')
        kwargs = dict((k, v) for k, v in config.items() if k not in _redis_kwargs_exclusions)
        if 'socket_timeout' not in kwargs:
            kwargs['socket_timeout'] = _DEFAULT_SOCKET_TIMEOUT
        if 'socket_connect_timeout' not in kwargs:
            kwargs['socket_connect_timeout'] = _DEFAULT_SOCKET_TIMEOUT
        if 'socket_keepalive' not in kwargs:
            kwargs['socket_keepalive'] = 1
        if 'socket_keepalive_options' not in kwargs:
            kwargs['socket_keepalive_options'] = _TCP_KEEP_ALIVE_OPTIONS
        if kwargs.pop('blocking_pool', False):
            if 'blocking_pool_timeout' in kwargs:
                kwargs['timeout'] = kwargs.pop('blocking_pool_timeout')
            else:
                kwargs['timeout'] = _DEFAULT_REDIS_BLOCKING_POOL_TIMEOUT
            connection_pool = redis.BlockingConnectionPool(**kwargs)
        else:
            connection_pool = redis.ConnectionPool(**kwargs)

        self._client = redis.Redis(connection_pool=connection_pool)

    def dump_object(self, value):
        """Dumps an object into a string for redis.  By default it serializes
        integers as regular string and pickle dumps everything else.
        """
        t = type(value)
        if t in _integer_types:
            return str(value).encode('ascii')
        return b'!' + pickle.dumps(value)

    def load_object(self, value):
        """The reversal of :meth:`dump_object`.  This might be callde with
        None.
        """
        if value is None:
            return None
        if value.startswith(b'!'):
            try:
                return pickle.loads(value[1:])
            except pickle.PickleError:
                return None
        try:
            return int(value)
        except ValueError:
            return value

    def get(self, key):
        return self.load_object(self._client.get(self.key_prefix + key))

    def get_list(self, keys):
        if self.key_prefix:
            keys = [self.key_prefix + key for key in keys]
        return [self.load_object(x) for x in self._client.mget(keys)]

    def get_many(self, keys):
        if self.key_prefix:
            query_keys = [self.key_prefix + key for key in keys]
        else:
            query_keys = keys
        values_list = self._client.mget(query_keys)
        values = {}
        for i in xrange(len(keys)):
            value = values_list[i]
            if value is not None:
                values[keys[i]] = self.load_object(value)
        return values

    def set(self, key, value, timeout=None, noreply=False):
        if timeout is None:
            timeout = self.default_timeout
        dump = self.dump_object(value)
        if timeout == 0:
            return self._client.set(name=self.key_prefix + key,value=dump)
        else:
            return self._client.setex(name=self.key_prefix + key, value=dump, time=timeout)

    def add(self, key, value, timeout=None, noreply=False):
        if timeout is None:
            timeout = self.default_timeout
        dump = self.dump_object(value)
        result = self._client.setnx(name=self.key_prefix + key, value=dump)
        if result and timeout != 0:
            result = self._client.expire(name=self.key_prefix + key, time=timeout)
        return result

    def set_many(self, data, timeout=None, noreply=False):
        if timeout is None:
            timeout = self.default_timeout
        pipe = self._client.pipeline()
        for key, value in _items(data):
            dump = self.dump_object(value)
            if timeout == 0:
                pipe.set(name=self.key_prefix + key, value=dump)
            else:
                pipe.setex(name=self.key_prefix + key, value=dump, time=timeout)
        return pipe.execute()

    def delete(self, key, noreply=False):
        self._client.delete(self.key_prefix + key)
        return True

    def delete_many(self, keys, noreply=False):
        if not keys:
            return True
        if self.key_prefix:
            keys = [self.key_prefix + key for key in keys]
        self._client.delete(*keys)
        return True

    def clear(self):
        status = False
        if self.key_prefix:
            keys = self._client.keys(self.key_prefix + '*')
            if keys:
                status = self._client.delete(*keys)
        else:
            status = self._client.flushdb()
        return status

    def incr(self, key, delta=1, noreply=False):
        return self._client.incr(name=self.key_prefix + key, amount=delta)

    def decr(self, key, delta=1, noreply=False):
        return self._client.decr(name=self.key_prefix + key, amount=delta)

    def hgetall(self, key):
        value = self._client.hgetall(self.key_prefix + key)
        if value is not None:
            for field in value:
                value[field] = self.load_object(value[field])
        return value

    def hget(self, key, field):
        return self.load_object(self._client.hget(self.key_prefix + key, field))

    def hset(self, key, field, value, timeout=None, noreply=False):
        self._client.hset(self.key_prefix + key, field, self.dump_object(value))
        return True

    def hdel(self, key, field, noreply=False):
        self._client.hdel(self.key_prefix + key, field)
        return True

class RawRedisCache(RedisCache):
    """Same cache client as RedisCache, but only support string value.
    """

    def __init__(self, config):
        RedisCache.__init__(self, config)

    def dump_object(self, value):
        if not isinstance(value, str):
            raise Exception('raw_redis_unsupport_value_type:' + str(type(value)))
        return value

    def load_object(self, value):
        return value

class _SsdbConnectionPool(object):

    def __init__(self, connection):
        self._connection = connection

    def get_connection(self, command_name, *keys, **options):
        return self._connection

    def release(self, connection):
        return

class SsdbCache(BaseCache):
    """Uses the SSDB key-value store as a cache backend.

    :param host: address of the SSDB server.
    :param port: port number on which SSDB server listens for connections.
    :param default_timeout: the default timeout that is used if no timeout is
                            specified on :meth:`~BaseCache.set`.
                            defaults to 0, means never expire.
    :param key_prefix: A prefix that should be added to all keys.

    Any additional keyword arguments will be passed to ``ssdb.Connection``.
    """

    def __init__(self, config):
        BaseCache.__init__(self, config)
        self.default_timeout = config.get('default_timeout', 0)
        self.key_prefix = config.get('key_prefix', '')
        try:
            import ssdb
            #patch ssdb
            import six
            import datetime
            ssdb.connection.iteritems = six.iteritems
            def expire(self, name, ttl):
                if isinstance(ttl, datetime.timedelta):
                    ttl = ttl.seconds + ttl.days * 24 * 3600
                return self.execute_command('expire', name, ttl)
            ssdb.SSDB.expire = expire
        except ImportError:
            raise RuntimeError('no ssdb module found')
        kwargs = dict((k, v) for k, v in config.items() if k not in _redis_kwargs_exclusions)
        if 'socket_timeout' not in kwargs:
            kwargs['socket_timeout'] = _DEFAULT_SOCKET_TIMEOUT
        if 'socket_connect_timeout' not in kwargs:
            kwargs['socket_connect_timeout'] = _DEFAULT_SOCKET_TIMEOUT
        if 'socket_keepalive' not in kwargs:
            kwargs['socket_keepalive'] = 1
        if 'socket_keepalive_options':
            kwargs['socket_keepalive_options'] = _TCP_KEEP_ALIVE_OPTIONS
        connection_pool = _SsdbConnectionPool(ssdb.Connection(**kwargs))
        self._client = ssdb.SSDB(connection_pool=connection_pool)

    def get(self, key):
        if self.key_prefix:
            key = self.key_prefix + key
        return self._client.get(key)

    def get_many(self, keys):
        if self.key_prefix:
            keys = [self.key_prefix + key for key in keys]
        values = self._client.multi_get(*keys)
        if self.key_prefix:
            values = dict([(key[len(self.key_prefix):], value) for key, value in _iteritems(values)])
        return values

    def set(self, key, value, timeout=None, noreply=False):
        if self.key_prefix:
            key = self.key_prefix + key
        if timeout is None:
            timeout = self.default_timeout
        if timeout == 0:
            return self._client.set(key, value)
        else:
            return self._client.setx(key, value, timeout)

    def add(self, key, value, timeout=None, noreply=False):
        if self.key_prefix:
            key = self.key_prefix + key
        if timeout is None:
            timeout = self.default_timeout
        result = self._client.setnx(key, value)
        if result and timeout != 0:
            result = self._client.expire(key, timeout)
        return result

    def set_many(self, data, timeout=None, noreply=False):
        if timeout is None:
            timeout = self.default_timeout
        if self.key_prefix:
            new_data = {}
            for key in data:
                new_data[self.key_prefix + key] = data[key]
            data = new_data
        result = self._client.multi_set(**data)
        if result and timeout != 0:
            for key in data:
                result = self._client.expire(key, timeout)
        return result

    def delete(self, key, noreply=False):
        if self.key_prefix:
            key = self.key_prefix + key
        self._client.delete(key)
        return True

    def delete_many(self, keys, noreply=False):
        if not keys:
            return True
        if self.key_prefix:
            keys = [self.key_prefix + key for key in keys]
        self._client.multi_del(*keys)
        return True

    def clear(self):
        keys = self._client.keys('', '')
        if self.key_prefix:
            keys = [key for key in keys if key.startswith(self.key_prefix)]
        status = self._client.delete(*keys)
        return status

    def incr(self, key, delta=1, noreply=False):
        if self.key_prefix:
            key = self.key_prefix + key
        return self._client.incr(self.key_prefix + key, delta)

    def decr(self, key, delta=1, noreply=False):
        if self.key_prefix:
            key = self.key_prefix + key
        return self._client.decr(self.key_prefix + key, delta)

    def hgetall(self, key):
        if self.key_prefix:
            key = self.key_prefix + key
        return self._client.hgetall(key)

    def hget(self, key, field):
        if self.key_prefix:
            key = self.key_prefix + key
        return self._client.hget(key, field)

    def hset(self, key, field, value, timeout=None, noreply=False):
        if self.key_prefix:
            key = self.key_prefix + key
        return self._client.hset(key, field, value)

    def hdel(self, key, field, noreply=False):
        if self.key_prefix:
            key = self.key_prefix + key
        return self._client.hdel(key, field)


class ReplicationCache(BaseCache):

    def __init__(self, config):
        BaseCache.__init__(self, config)
        self._client = []
        self._primary_cache = None
        self._id = config['id']
        primary = config.get('primary')
        caches = config['caches']
        for index, name in enumerate(caches.keys()):
            self._client.append(create_cache_client(name, caches[name]))
            if primary is not None and primary == name:
                self._primary_cache = self._client[index]

    def get_client(self):
        if self._primary_cache is not None:
            return self._primary_cache
        if len(self._client) == 0:
            return None
        return random.choice(self._client)

    def set(self, key, value, timeout=None, noreply=False):
        result = True
        for client in self._client:
            if not client.set(key, value, timeout, noreply):
                result = False
        return result

    def add(self, key, value, timeout=None, noreply=False):
        result = True
        for client in self._client:
            if not client.add(key, value, timeout, noreply):
                result = False
        return result

    def set_many(self, data, timeout=None, noreply=False):
        result = True
        for client in self._client:
            if not client.set_many(data, timeout, noreply):
                result = False
        return result

    def get(self, key):
        client = self.get_client()
        if client is None:
            return None
        return client.get(key)

    def get_list(self, keys):
        client = self.get_client()
        if client is None:
            return None
        return client.get_list(keys)

    def get_many(self, keys):
        client = self.get_client()
        if client is None:
            return None
        return client.get_many(keys)

    def delete(self, key, noreply=False):
        result = True
        for client in self._client:
            if not client.delete(key, noreply):
                result = False
        return result

    def delete_many(self, keys, noreply=False):
        result = True
        for client in self._client:
            if not client.delete_many(keys, noreply):
                result = False
        return result

    def clear(self):
        result = True
        for client in self._client:
            if not client.clear():
                result = False
        return result

    def incr(self, key, delta=1, noreply=False):
        result = None
        for client in self._client:
            result = client.incr(key, delta, noreply)
            if result is None:
                return None
        return result

    def decr(self, key, delta=1, noreply=False):
        result = None
        for client in self._client:
            result = client.decr(key, delta, noreply)
            if result is None:
                return None
        return result

    def hgetall(self, key):
        client = self.get_client()
        if client is None:
            return None
        return client.hgetall(key)

    def hget(self, key, field):
        client = self.get_client()
        if client is None:
            return None
        return client.hget(key, field)

    def hset(self, key, field, value, timeout=None, noreply=False):
        result = True
        for client in self._client:
            if not client.hset(key, field, value, timeout, noreply):
                result = False
        return result

    def hdel(self, key, field, noreply=False):
        result = True
        for client in self._client:
            if not client.hdel(key, field, noreply):
                result = False
        return result


class DistributionCache(BaseCache):
    """A cache distributior based on cache key.

    :param method: support conhash / mod / div. default is conhash
        conhash: children should have property replica, default replica is 32
        mod: require key_regex, factor. client_id = key % factor
        div: require key_regex, factor. client_id = key / factor
    :param key_regex: regex used to extract base key from cache key to caculate client_id.
        the first regex group matched in cache key will be used as base key.
    :param factor: divisor in mod or div method..
    """

    def __init__(self, config):
        BaseCache.__init__(self, config)
        self._client = {}
        self._id = config['id']
        self._method = config.get('method')
        if self._method == 'mod' or self._method == 'div':
            self._init_method_mod_div(config)
        else:
            self._init_method_conhash(config)

    def _init_method_conhash(self, config):
        self._conhash = ConHash()
        caches = config['caches']
        for index, name in enumerate(caches.keys()):
            self._client[name] = create_cache_client(name, caches[name])
            self._conhash.add_node(str(name), caches[name].get('replica', 32), index)
        self.get_client_by_key = self._get_client_by_key_conhash

    def _get_client_by_key_conhash(self, key):
        client_id = self._conhash.lookup(str(key))
        return self._client.get(client_id)

    def _init_method_mod_div(self, config):
        self._regex = re.compile(config['key_regex'])
        self._factor = config['factor']
        caches = config['caches']
        for name in caches:
            self._client[int(name)] = create_cache_client(name, caches[name])
        if self._method == 'mod':
            self.get_client_by_key = self._get_client_by_key_mod
        else:
            self.get_client_by_key = self._get_client_by_key_div

    def _get_client_by_key_mod(self, key):
        try:
            client_id = int(self._regex.match(key).group(1)) % self._factor
        except:
            return None
        return self._client.get(client_id)

    def _get_client_by_key_div(self, key):
        try:
            client_id = int(self._regex.match(key).group(1)) / self._factor
        except:
            return None
        return self._client.get(client_id)

    def set(self, key, value, timeout=None, noreply=False):
        client = self.get_client_by_key(key)
        if client is None:
            return False
        return client.set(key, value, timeout, noreply)

    def add(self, key, value, timeout=None, noreply=False):
        client = self.get_client_by_key(key)
        if client is None:
            return False
        return client.add(key, value, timeout, noreply)

    def set_many(self, data, timeout=None, noreply=False):
        query_table = defaultdict(dict)
        for key in data:
            client = self.get_client_by_key(key)
            if client is None:
                return False
            query_table[client][key] = data[key]
        for client, query in _iteritems(query_table):
            if not client.set_many(query, timeout, noreply):
                return False
        return True

    def get(self, key):
        client = self.get_client_by_key(key)
        if client is None:
            return None
        return client.get(key)

    def get_many(self, keys):
        query_table = defaultdict(list)
        for key in keys:
            client = self.get_client_by_key(key)
            if client is None:
                return None
            query_table[client].append(key)
        values = {}
        for client, query in _iteritems(query_table):
            result = client.get_many(query)
            if result is None:
                return None
            values.update(result)
        return values

    def delete(self, key, noreply=False):
        client = self.get_client_by_key(key)
        if client is None:
            return False
        return client.delete(key, noreply)

    def delete_many(self, keys, noreply=False):
        query_table = defaultdict(list)
        for key in keys:
            client = self.get_client_by_key(key)
            if client is None:
                return False
            query_table[client].append(key)
        for client, query in _iteritems(query_table):
            if not client.delete_many(query, noreply):
                return False
        return True

    def clear(self):
        result = True
        for client in self._client.values():
            if not client.clear():
                result = False
        return result

    def incr(self, key, delta=1, noreply=False):
        client = self.get_client_by_key(key)
        if client is None:
            return None
        return client.incr(key, delta, noreply)

    def decr(self, key, delta=1, noreply=False):
        client = self.get_client_by_key(key)
        if client is None:
            return None
        return client.decr(key, delta, noreply)

    def hgetall(self, key):
        client = self.get_client_by_key(key)
        if client is None:
            return None
        return client.hgetall(key)

    def hget(self, key, field):
        client = self.get_client_by_key(key)
        if client is None:
            return None
        return client.hget(key, field)

    def hset(self, key, field, value, timeout=None, noreply=False):
        client = self.get_client_by_key(key)
        if client is None:
            return None
        return client.hset(key, field, value, timeout, noreply)

    def hdel(self, key, field, noreply=False):
        client = self.get_client_by_key(key)
        if client is None:
            return None
        return client.hdel(key, field, noreply)


_cache_classes = {
    'null': NullCache,
    'memcached': MemcachedCache,
    'pymemcached': PyMemcachedCache,
    'redis': RedisCache,
    'rawredis': RawRedisCache,
    'ssdb': SsdbCache,
    'replication': ReplicationCache,
    'distribution': DistributionCache
}

def create_cache_client(cache_id, config):
    cache_type = config['type']
    config['id'] = cache_id
    if cache_type not in _cache_classes:
        raise RuntimeError('unknown_cache_type: %s' % cache_type)
    return _cache_classes[cache_type](config)

_cache_clients = {}
_default_client = None

def init_cache(config):
    """ init cache clients
    :param config:
    {
        'main':  {
            'type': 'memcached',
            'host': '127.0.0.1',
            'port': 11211,
            'default_timeout': 7 * 24 * 60 * 60,
            'key_prefix': 'test.',
            'default': True,
        },
        'test':  {
            'type': 'redis',
            'host': '127.0.0.1',
            'port': 6379,
            'default_timeout': 7 * 24 * 60 * 60,
            'key_prefix': 'test.',
        },
        'null':  {
            'type': 'null',
        },
        'replication': {
            'type': 'replication',
            'primary': 'replication.main',
            'caches': {
                'replication.main': {
                    'type': 'distribution',
                    'caches': {
                        'distribution.1': {
                            'type': 'memcached',
                            'host': '127.0.0.1',
                            'port': 11211,
                            'default_timeout': 7 * 24 * 60 * 60,
                            'key_prefix': 'test',
                            'replica': 32
                        },
                        'distribution.2': {
                            'type': 'memcached',
                            'host': '127.0.0.1',
                            'port': 11212,
                            'default_timeout': 7 * 24 * 60 * 60,
                            'key_prefix': 'test',
                            'replica': 32
                        }
                    }
                },
                'replication.bak': {
                    'type': 'distribution',
                    'method': 'mod',
                    'key_regex': r'\w+\.(\d+)',
                    'factor': 2,
                    'caches': {
                        '0': {
                            'type': 'memcached',
                            'host': '127.0.0.1',
                            'port': 11211,
                            'default_timeout': 7 * 24 * 60 * 60,
                            'key_prefix': 'test'
                        },
                        '1': {
                            'type': 'memcached',
                            'host': '127.0.0.1',
                            'port': 11212,
                            'default_timeout': 7 * 24 * 60 * 60,
                            'key_prefix': 'test'
                        }
                    }
                }
            }
        }
    }
    null: default_timeout
    memcached: host, port, default_timeout, key_prefix
    pymemcached: host, port, default_timeout, key_prefix. this client supports noreply.
    redis: host, port, password, db, default_timeout, key_prefix
    rawredis: host, port, password, db, default_timeout, key_prefix
    ssdb: host, port, default_timeout, key_prefix
    distribution: caches, method, key_regex, factor
    replication: caches, primary
    there should be one instance with default=True
    :return: None
    """
    global _cache_clients
    global _default_client
    for name in config:
        item = config[name]
        client = create_cache_client(name, item)
        _cache_clients[name] = client
        if item.get('default'):
            _default_client = client

def get_cache(name=None):
    if name is None:
        return _default_client
    return _cache_clients.get(name)

def get(key):
    return _default_client.get(key)

def delete(key, noreply=False):
    return _default_client.delete(key, noreply)

def get_list(keys):
    return _default_client.get_list(keys)

def get_many(keys):
    return _default_client.get_many(keys)

def set(key, value, timeout=None, noreply=False):
    return _default_client.set(key, value, timeout, noreply)

def add(key, value, timeout=None, noreply=False):
    return _default_client.add(key, value, timeout, noreply)

def set_many(data, timeout=None, noreply=False):
    return _default_client.set_many(data, timeout, noreply)

def delete_many(keys, noreply=False):
    return _default_client.delete_many(keys, noreply)

def clear():
    return _default_client.clear()

def incr(key, delta=1, noreply=False):
    return _default_client.incr(key, delta, noreply)

def decr(key, delta=1, noreply=False):
    return _default_client.decr(key, delta, noreply)

def hgetall(key):
    return _default_client.hgetall(key)

def hget(key, field):
    return _default_client.hget(key, field)

def hset(key, field, value, timeout=None, noreply=False):
    return _default_client.hset(key, field, value, timeout, noreply)

def hdel(key, field, noreply=False):
    return _default_client.hdel(key, field, noreply)



import psutil

def kill_process_tree(pid):
    try:
        main_proc = psutil.Process(pid)
    except (psutil.NoSuchProcess, psutil.AccessDenied) as ex:
        return
    processes = main_proc.children(recursive=True)
    processes.append(main_proc)
    for proc in processes:
        proc.terminate()
    psutil.wait_procs(processes)




# coding: utf-8

# import platform

# if platform.system().lower() == 'darwin':
#     from sakura.cpplib_osx import ConHash
# else:
#     from sakura.cpplib import ConHash
import sys
from hashlib import md5

_PY2 = sys.version_info[0] == 2


if _PY2:
    pass
else:
    long = int
    xrange = range


class ConHash(object):

    def __init__(self, nodes=None, replicas=3):
        """ nodes 是字符串的列表
            replicas 表示有多少个虚拟节点
        """
        self.replicas = replicas
        self.ring = dict()
        self._sorted_keys = []  # 有序的 hash node 队列
        self.lookup = self.get_node
        if nodes:
            for node in nodes:
                self.add_node(node)

    def add_node(self, node, *args):
        """将 node 节点加入到哈希环中
        """
        for i in xrange(0, self.replicas):
            key = self.gen_key('%s:%s' % (node, i))
            self.ring[key] = node
            self._sorted_keys.append(key)
        self._sorted_keys.sort()

    def remove_node(self, node):
        for i in xrange(0, self.replicas):
            key = self.gen_key('%s:%s' % (node, i))
            del self.ring[key]
            self._sorted_keys.remove(key)

    def get_node(self, string_key):
        """根据 string_key 找到相应的 node 节点
        """
        return self.get_node_pos(string_key)[0]

    def get_node_pos(self, string_key):
        """根据 string key 找到 node ,如果没找到，则返回第一个节点
        """
        if not self.ring:
            return None, None
        key = self.gen_key(string_key)
        nodes = self._sorted_keys
        for i in xrange(0, len(nodes)):
            node = nodes[i]
            if key <= node:
                return self.ring[node], i
        return self.ring[nodes[0]], 0

    def get_nodes(self, string_key):
        """ 获取 所有可用的 nodes 节点
        """
        if not self.ring:
            yield None, None
        node, pos = self.get_node_pos(string_key)
        # 先返回 string_key pos 节点后的 nodes
        for key in self._sorted_keys[pos:]:
            yield self.ring[key]
        # 然后循环返回整个列表
        while True:
            for key in self._sorted_keys:
                yield self.ring[key]

    def gen_key(self, key):
        """生成 key
        """
        m = md5.new()
        m.update(key)
        return long(m.hexdigest(), 16)


# _*_ coding:utf8 _*_
'''
Common logger library

[Config]
log_dir: log files directory. default './log'
sentry_dsn: sentry dsn. default no sentry
rollover_when: rollover interval. default 'MIDNIGHT'
    S - Seconds
    M - Minutes
    H - Hours
    D - Days
    MIDNIGHT - roll over at midnight
rollover_backup_count: how many backup log files are kept. default 30
    if rollover_backup_count = 0, all log files are kept.
    if rollover_backup_count > 0, when rollover is done, no more than rollover_backup_count files are kept - the oldest ones are deleted.

[Normal Python Program]
# config.py
# Add LOGGER_CONFIG
LOGGER_CONFIG = {
    'log_dir': './log',
    'sentry_dsn': 'http://xxxxxxxx',
}
#if no config.py, will use './log' as log_dir

[Integrate Django]
# settings.py
# Add LOGGER_CONFIG
LOGGER_CONFIG = {
    'log_dir': './log',
    'sentry_dsn': 'http://xxxxxxxx',
    'sentry_project_release': '1860da72a8c86bf4832a598410186ae668bbaf20',  # used by sentry to track which git release does error belong to
}
# Add 'raven.contrib.django.raven_compat' in INSTALLED_APPS
INSTALLED_APPS = (
    ...,
    'raven.contrib.django.raven_compat',
)
# Add 'raven.contrib.django.middleware.SentryLogMiddleware' in MIDDLEWARE_CLASSES
MIDDLEWARE_CLASSES = (
    ...,
    'raven.contrib.django.middleware.SentryLogMiddleware',
)
'''
#
# log = None
#
# try:
#     from django.utils.log import AdminEmailHandler
#
#     class ErrorEmailHandler(AdminEmailHandler):
#         def format_subject(self, subject):
#             pos_list = filter(lambda x:x>0, [subject.find(c) for c in '|\r\n'])
#             if pos_list:
#                     return subject[:min(pos_list)]
#             else:
#                     return subject
# except:
#     pass
#
# def _log_record_exception(func):
#     def _func(self):
#         try:
#             return func(self)
#         except:
#             log.exception('log_exception|thread=%s:%s,file=%s:%s,func=%s:%s,log=%s',
#                 self.process, self.thread, self.filename, self.lineno, self.module, self.funcName, self.msg)
#             raise
#     return _func
#
# def append_exc(func):
#     def _append_exc(*args,**kwargs):
#         if 'exc_info' not in kwargs:
#             kwargs['exc_info'] = True
#         return func(*args,**kwargs)
#     return _append_exc
#
# def init_logger(log_dir=None, sentry_dsn=None, sentry_project_release=None, rollover_when='MIDNIGHT', rollover_backup_count=30):
#
#     if log_dir is None:
#         log_dir = '/var/log/sakura'
#
#     import os
#     import sys
#     log_dir = os.path.abspath(log_dir)
#     if log_dir and not os.path.exists(log_dir):
#         os.mkdir(log_dir)
#
#     logger_config = {
#         'version': 1,
#         'disable_existing_loggers': True,
#         'formatters': {
#             'standard': {
#                 'format': '%(asctime)s.%(msecs)03d|%(levelname)s|%(process)d:%(thread)d|%(filename)s:%(lineno)d|%(module)s.%(funcName)s|%(message)s',
#                 'datefmt': '%Y-%m-%d %H:%M:%S',
#             },
#             'short' : {
#                 'format': '%(asctime)s.%(msecs)03d|%(levelname)s|%(message)s',
#                 'datefmt': '%Y-%m-%d %H:%M:%S',
#             },
#             'data' : {
#                 'format': '%(asctime)s.%(msecs)03d|%(message)s',
#                 'datefmt': '%Y-%m-%d %H:%M:%S',
#             },
#         },
#         'handlers': {
#             'file_fatal': {
#                 'level':'CRITICAL',
#                 'class':'sakura.loggingmp.MPTimedRotatingFileHandler',
#                 'filename': os.path.join(log_dir, 'fatal.log').replace('\\','/'),
#                 'when': rollover_when,
#                 'backupCount': rollover_backup_count,
#                 'formatter':'standard',
#             },
#             'file_error': {
#                 'level':'WARNING',
#                 'class':'sakura.loggingmp.MPTimedRotatingFileHandler',
#                 'filename': os.path.join(log_dir, 'error.log').replace('\\','/'),
#                 'when': rollover_when,
#                 'backupCount': rollover_backup_count,
#                 'formatter':'standard',
#             },
#             'file_info': {
#                 'level':'DEBUG',
#                 'class':'sakura.loggingmp.MPTimedRotatingFileHandler',
#                 'filename': os.path.join(log_dir, 'info.log').replace('\\','/'),
#                 'when': rollover_when,
#                 'backupCount': rollover_backup_count,
#                 'formatter':'short',
#             },
#             'file_data': {
#                 'level':'DEBUG',
#                 'class':'sakura.loggingmp.MPTimedRotatingFileHandler',
#                 'filename': os.path.join(log_dir, 'data.log').replace('\\','/'),
#                 'when': rollover_when,
#                 'backupCount': rollover_backup_count,
#                 'formatter':'data',
#             },
#         },
#         'loggers': {
#             'main': {
#                 'handlers': ['file_fatal', 'file_error', 'file_info'],
#                 'level': 'DEBUG',
#                 'propagate': True,
#             },
#             'data': {
#                 'handlers': ['file_data'],
#                 'level': 'DEBUG',
#                 'propagate': True,
#             },
#             'django.request': {
#                 'handlers': ['file_fatal', 'file_error', 'file_info'],
#                 'level': 'ERROR',
#                 'propagate': True,
#             },
#             'tornado.access': {
#                 'handlers': ['file_data'],
#                 'level': 'DEBUG',
#                 'propagate': True,
#             },
#             'tornado.application': {
#                 'handlers': ['file_fatal', 'file_error', 'file_info'],
#                 'level': 'DEBUG',
#                 'propagate': True,
#             },
#             'tornado.general': {
#                 'handlers': ['file_fatal', 'file_error', 'file_info'],
#                 'level': 'DEBUG',
#                 'propagate': True,
#             },
#         }
#     }
#
#     is_django_app = False
#     is_debug = False
#     is_test = False
#     try:
#         from django.conf import settings
#         is_django_app = settings.configured
#         if is_django_app:
#             is_debug = settings.DEBUG
#             is_test = 'TEST' in dir(settings) and settings.TEST
#     except:
#         pass
#     if not is_django_app:
#         import config
#         is_debug = 'DEBUG' in dir(config) and config.DEBUG
#         is_test = 'TEST' in dir(config) and config.TEST
#
#     if is_debug:
#         logger_config['handlers']['file_debug'] = {
#             'level':'DEBUG',
#             'class':'sakura.loggingmp.MPTimedRotatingFileHandler',
#             'filename': os.path.join(log_dir, 'debug.log').replace('\\','/'),
#             'when': rollover_when,
#             'backupCount': rollover_backup_count,
#             'formatter':'short',
#         }
#         logger_config['loggers']['django.db.backends'] = {
#             'handlers': ['file_debug'],
#             'level': 'DEBUG',
#             'propagate': True,
#         }
#     elif not is_test:
#         loggers = logger_config['loggers']
#         for logger_item in loggers:
#             if loggers[logger_item]['level'] == 'DEBUG':
#                 loggers[logger_item]['level'] = 'INFO'
#
#     if not is_debug and sentry_dsn is not None:
#         try:
#             import raven
#             if is_django_app:
#                 logger_config['handlers']['sentry'] = {
#                     'level':'WARNING',
#                     'class':'raven.contrib.django.raven_compat.handlers.SentryHandler',
#                     'formatter':'short',
#                 }
#                 settings.RAVEN_CONFIG = {'dsn': sentry_dsn, 'release': sentry_project_release}
#             else:
#                 logger_config['handlers']['sentry'] = {
#                     'level':'WARNING',
#                     'class':'raven.handlers.logging.SentryHandler',
#                     'dsn': sentry_dsn,
#                     #'auto_log_stacks': True,
#                     'formatter':'short',
#                     'release': sentry_project_release,
#                 }
#             logger_config['loggers']['django.request']['handlers'].append('sentry')
#             logger_config['loggers']['main']['handlers'].append('sentry')
#         except:
#             pass
#
#     work_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)), '../')
#     recover_path = False
#     if work_dir not in sys.path:
#         sys.path.append(work_dir)
#         recover_path = True
#
#     import logging
#     try:
#         import logging.config
#         logging.config.dictConfig(logger_config)
#     except:
#         import loggerconfig
#         loggerconfig.dictConfig(logger_config)
#
#     if recover_path:
#         sys.path.remove(work_dir)
#
#     global log
#     log = logging.getLogger('main')
#     log.exception = append_exc(log.error)
#     log.assertion = log.critical
#     log.data = logging.getLogger('data').info
#     logging.LogRecord.getMessage = _log_record_exception(logging.LogRecord.getMessage)
#
# #try init log
# def try_init_logger():
#     try:
#         from django.conf import settings
#
#         setting_keys = dir(settings)
#         if 'LOGGER_CONFIG' in setting_keys:
#             init_logger(**settings.LOGGER_CONFIG)
#         elif 'LOGGING' in setting_keys and settings.LOGGING:
#             import logging
#             global log
#             log = logging.getLogger('main')
#             log.exception = append_exc(log.error)
#             log.data = logging.getLogger('data').info
#         else:
#             init_logger()
#     except:
#         try:
#             import config
#             init_logger(**config.LOGGER_CONFIG)
#         except:
#             try:
#                 init_logger()
#             except:
#                 pass
#
# if log is None:
#     try_init_logger()


import logging.handlers
from sakura import config

class Logger:
    logger = None

    levels = {"n": logging.NOTSET,  # 不设置    0
              "d": logging.DEBUG,  # 调试信息  10
              "i": logging.INFO,  # 有用信息  20
              "w": logging.WARN,  # 警告信息  30
              "e": logging.ERROR,  # 错误信息  40
              "c": logging.CRITICAL}  # 严重错误信息 50

    log_level = "d"
    log_file = "/var/log/sakura/sakura.log"
    log_dict = getattr(config, "LOGGER_CONFIG", {})
    if log_dict:
        log_file = log_dict.get("log_file", log_file)
    log_max_byte = 10 * 1024 * 1024
    log_backup_count = 5

    @staticmethod
    def getLogger():
        if Logger.logger is not None:
            return Logger.logger

        Logger.logger = logging.Logger("Logger")
        log_handler = logging.handlers.RotatingFileHandler(filename=Logger.log_file,
                                                           maxBytes=Logger.log_max_byte,
                                                           backupCount=Logger.log_backup_count)
        log_fmt = logging.Formatter('%(asctime)s %(levelname)s %(filename)s %(funcName)s %(message)s',
                                    '%Y-%m-%d %H:%M:%S')
        log_handler.setFormatter(log_fmt)
        Logger.logger.addHandler(log_handler)
        Logger.logger.setLevel(Logger.levels.get(Logger.log_level))
        return Logger.logger


def append_exc(func):
    def _append_exc(*args,**kwargs):
        if 'exc_info' not in kwargs:
            kwargs['exc_info'] = True
        return func(*args,**kwargs)
    return _append_exc


log = Logger.getLogger()
log.data = logging.getLogger('data').info
log.exception = append_exc(log.error)

del Logger



try:
    import simplejson as json
except:
    import json

def to_json(data, ensure_ascii=False):
    return json.dumps(data, ensure_ascii=ensure_ascii, separators=(',', ':'))

def from_json(s):
    return json.loads(s)

def from_json_safe(s):
    try:
        return json.loads(s)
    except:
        return None

_slash_escape = '\\/' in to_json('/')

def to_json_html_safe(obj, **kwargs):
    rv = to_json(obj, **kwargs)    \
        .replace(u'<', u'\\u003c')    \
        .replace(u'>', u'\\u003e')    \
        .replace(u'&', u'\\u0026')    \
        .replace(u"'", u'\\u0027')    \
        .replace(u'\u2028', u'\\u2028')    \
        .replace(u'\u2029', u'\\u2029')
    if _slash_escape:
        rv = rv.replace('\\/', '/')
    return rv



import inspect
import os
import copy
import time
from contextlib import contextmanager
from functools import wraps
from sakura.logger import log
from sakura.convert import *
import crypt


class ClassPropertyMetaClass(type):
    def __setattr__(self, key, value):
        if key in self.__dict__:
            obj = self.__dict__.get(key)
            if obj and type(obj) is ClassPropertyDescriptor:
                return obj.__set__(self, value)
        super(ClassPropertyMetaClass, self).__setattr__(key, value)

class ClassPropertyDescriptor(object):
    def __init__(self, fget, fset=None):
        self.fget = fget
        self.fset = fset

    def __get__(self, obj, klass=None):
        if klass is None:
            klass = type(obj)
        return self.fget.__get__(obj, klass)()

    def __set__(self, obj, value):
        if not self.fset:
            raise AttributeError("can't set attribute")
        if inspect.isclass(obj):
            type_ = obj
            obj = None
        else:
            type_ = type(obj)
        return self.fset.__get__(obj, type_)(value)

    def setter(self, func):
        if not isinstance(func, (classmethod, staticmethod)):
            func = classmethod(func)
        self.fset = func
        return self

def classproperty(func):
    """
        -- usage --
        1) simple read:
                ...
                @classproperty
                def MY_VALUE(cls):
                    return cls._MY_VALUE
                ...

        2) simple write: (write by instance)
                ...
                // add setter
                @MY_VALUE.setter
                def MY_VALUE(cls, value):
                    cls._MY_VALUE = value

        3) complete write: (write by instance or class)
                ...
                // add metaclass
                __metaclass__ = ClassPropertyMetaClass

    """
    if not isinstance(func, (classmethod, staticmethod)):
        func = classmethod(func)
    return ClassPropertyDescriptor(func)

class Object:
    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)

def dict_to_object(d):
    return Object(**d)

def create_object(**kwargs):
    return Object(**kwargs)

def find_first(f, seq):
    """Return first item in sequence where f(item) == True."""
    for item in seq:
        if f(item):
            return item
    return None

def get_timestamp():
    return int(time.time())

def find_str(text, prefix, suffix=None):
    start = text.find(prefix)
    if start < 0:
        return None
    start += len(prefix)
    if suffix is None:
        return text[start:]
    end = text.find(suffix, start)
    if end < 0:
        return None
    return text[start:end]

def truncate_unicode(text, max_length, encoding='utf-8', ending=u'...'):
    encoded_str = text.encode(encoding)
    if len(encoded_str) <= max_length:
        return text
    max_length -= len(ending)
    if max_length < 0:
        max_length = 0
    encoded_str = encoded_str[:max_length]
    return encoded_str.decode(encoding, 'ignore') + ending

def exception_safe(exception_return=None, keyword=None, return_filter=copy.copy):
    def _exception_safe(func):
        @wraps(func)
        def _func(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except:
                if keyword is None:
                    log.exception('%s_exception', func.__name__)
                else:
                    log.exception('%s_exception', keyword)
                if return_filter:
                    return return_filter(exception_return)
                else:
                    return exception_return
        return _func
    return _exception_safe

@contextmanager
def directory(path):
    current_dir = os.getcwd()
    os.chdir(path)
    try:
        yield
    finally:
        os.chdir(current_dir)

IS_DJANGO_APP = False
IS_FLASK_APP = False

try:
    from django.conf import settings
    if not settings.configured:
        raise Exception()
    from django_utils import *
    IS_DJANGO_APP = True
except:
    try:
        from flask_utils import *
        IS_FLASK_APP = True
    except:
        pass

def with_cache(cache_key_prefix, timeout=60*60):
    def _with_cache(func):
        @wraps(func)
        def _func(key):
            cache_key = cache_key_prefix + crypt.md5(key)
            cache_data = cache.get(cache_key)
            if cache_data is not None:
                if cache_data[0] == key:
                    return cache_data[1]
            data = func(key)
            cache_data = (key, data)
            cache.set(cache_key, cache_data, timeout)
            return data
        return _func
    return _with_cache



import random
import struct
import platform
from tornado.ioloop import IOLoop
from tornado.tcpserver import TCPServer
import socket
from sakura.logger import log
from sakura.simplequeue import SimpleQueue
from sakura.version_tools import byteshex, xrange


IPV6_V4_PREFIX = b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xff\xff'
IPV6_SIZE = 16
GTCP_PACKET_STUB_SIZE = 20
GTCP_CMD_SIZE = 1
GTCP_HEADER_SIZE = GTCP_CMD_SIZE + GTCP_PACKET_STUB_SIZE
GTCP_CMD_RELAY = b'\x00'
GTCP_CMD_NONE = b'\x01'
GTCP_CMD_ERROR = b'\x02'
GTCP_CMD_CONNECT = b'\x11'
GTCP_CMD_DISCONNECT = b'\x12'
GTCP_CMD_NOTIFY = b'\x13'


class TcpEndpoint(object):

    def __init__(self, client_id):
        self._client_id = client_id
        self._parsed = False

    def __str__(self):
        return self.address

    def _parse(self):
        if self._parsed:
            return
        self._raw_ip = self._client_id[:IPV6_SIZE]
        (self._port,) = struct.unpack_from('!H', self._client_id, IPV6_SIZE)
        if self._raw_ip.startswith(IPV6_V4_PREFIX):
            ipv4 = self._raw_ip[len(IPV6_V4_PREFIX):]
            (self._ip,) = struct.unpack('!I', ipv4)
            self._ip_str = socket.inet_ntoa(ipv4)
        else:
            self._ip = self._raw_ip
            self._ip_str = socket.inet_ntop(socket.AF_INET6, self._raw_ip)
        self._address = '%s:%u' % (self._ip_str, self._port)
        self._parsed = True

    @property
    def id(self):
        return self._client_id

    @property
    def client_id(self):
        return self._client_id

    @property
    def address(self):
        self._parse()
        return self._address

    @property
    def ip(self):
        self._parse()
        return self._ip

    @property
    def port(self):
        self._parse()
        return self._port


class WorkerTask(object):
    def __init__(self, client, cmd, packet=b''):
        self.client = client
        self.cmd = cmd
        self.packet = packet


class GTcpConnection(object):

    HEADER_SIZE = 4

    def __init__(self, stream, address, config, on_packet=None, on_close=None, endian="!"):
        self._stream = stream
        self._address = address
        self._endian = endian
        if stream.socket.family == socket.AF_INET6:
            self._remote_ip = socket.inet_pton(socket.AF_INET6, address[0])
        else:
            self._remote_ip = IPV6_V4_PREFIX + socket.inet_aton(address[0])
        self._remote_port = self._address[1]
        self._remote_address = '%s:%d' % address
        self._config = config
        if config.CONNECTION_ID_RANDOM_PADDING:
            padding = random.randint(0, 0xffff)
        else:
            padding = 0
        self._id = struct.pack('!16sHH', self._remote_ip, self._remote_port, padding)
        self._on_packet_callback = on_packet
        self._on_close_callback = on_close
        self._set_keep_alive()
        self._stream.set_close_callback(self._on_close)
        self._recv_header()

    @property
    def id(self):
        return self._id

    @property
    def remote_address(self):
        return self._remote_address

    def closed(self):
        return self._stream.closed()

    def _set_keep_alive(self):
        if not self._config.ENABLE_KEEP_ALIVE:
            return
        stream_socket = self._stream.socket
        stream_socket.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)
        if platform.system().lower() == 'linux':
            stream_socket.setsockopt(socket.SOL_TCP, socket.TCP_KEEPIDLE, self._config.KEEP_ALIVE_OPT['timeout'])
            stream_socket.setsockopt(socket.SOL_TCP, socket.TCP_KEEPINTVL, self._config.KEEP_ALIVE_OPT['interval'])
            stream_socket.setsockopt(socket.SOL_TCP, socket.TCP_KEEPCNT, self._config.KEEP_ALIVE_OPT['count'])

    def send(self, data):
        self._stream.write(data)

    def send_packet(self, packet):
        header = struct.pack('%sI' % self._endian, len(packet))
        self.send(header + packet)

    def close(self):
        self._stream.close()

    def _recv_header(self):
        self._stream.read_bytes(self.HEADER_SIZE, self._on_recv_header)

    def _on_recv_header(self, data):
        if len(data) != self.HEADER_SIZE:
            # log.error('tcp_conn_header_size_error|remote=%s,header=%s', self._remote_address, data.encode('hex'))
            log.error('tcp_conn_header_size_error|remote=%s,header=%s', self._remote_address, byteshex(data))
            self._stream.close()
            return
        (body_size,) = struct.unpack('%sI' % self._endian, data)
        log.info("offset size info: %s" % struct.unpack('<I', data))
        if body_size >= self._config.TCP_MAX_PACKET_SIZE:
            log.error('tcp_conn_body_size_overflow|remote=%s,size=%u', self._remote_address, body_size)
            self._stream.close()
            return
        self._stream.read_bytes(body_size, self._on_recv_body)

    def _on_recv_body(self, data):
        if self._on_packet_callback is not None:
            self._on_packet_callback(self, data)
        self._stream.read_bytes(self.HEADER_SIZE, self._on_recv_header)

    def _on_close(self):
        if self._on_close_callback is not None:
            self._on_close_callback(self)

class CallbackTcpServer(TCPServer):

    def __init__(self, on_connect=None, *args, **kwargs):
        self._on_connect_callback = on_connect
        super(CallbackTcpServer, self).__init__(*args, **kwargs)

    def handle_stream(self, stream, address):
        if self._on_connect_callback:
            self._on_connect_callback(stream, address)

class Processor(object):

    def __init__(self, id, config):
        self._id = id
        self._config = config
        self._client = None

    def run(self):
        random.seed()
        self.on_init()
        from sakura.gtcpclient import GTcpClient
        log.info('tcp_worker_start|id=%d', self._id)
        self._client = GTcpClient(self._config.WORK_ENDPOINT['address'], self._config.WORK_ENDPOINT['port'], 0)
        while True:
            try:
                request = self._client.receive()
                if request is None:
                    # log.warn('tcp_worker_lost_connection|client_id=%s,client=%s', self._client.id.encode('hex'), self._client.remote_address)
                    log.warn('tcp worker lost_connection|client_id=%s,client=%s', byteshex(self._id), self._client._address)
                    self._client.close()
                elif len(request) < GTCP_HEADER_SIZE:
                    # log.error('tcp_worker_request_packet_error|client_id=%s,client=%s,request=%s', self._client.id.encode('hex'), self._client.remote_address, request.encode('hex'))
                    log.error('tcp_worker_request_packet_error|client_id=%s,client=%s,request=%s', byteshex(self._id), self._client._address, byteshex(request))
                    self._client.close()
                else:
                    request_cmd = request[:GTCP_CMD_SIZE]
                    request_client = TcpEndpoint(request[GTCP_CMD_SIZE:GTCP_HEADER_SIZE])
                    reply_body = None
                    if request_cmd == GTCP_CMD_RELAY:
                        request_body = request[GTCP_HEADER_SIZE:]
                        reply_body = self.on_packet(request_client, request_body)
                    elif request_cmd == GTCP_CMD_CONNECT:
                        reply_body = self.on_client_connect(request_client)
                    elif request_cmd == GTCP_CMD_DISCONNECT:
                        self.on_client_disconnect(request_client)
                    if reply_body is None:
                        self._client.send(GTCP_CMD_NONE + request_client.client_id)
                    else:
                        self._client.send(GTCP_CMD_RELAY + request_client.client_id + reply_body)
            except Exception as ex:
                log.exception('tcp_worker_exception|id=%u,exception=%s', self._id, ex, exc_info=True)
                self._client.close()

    def send_packet(self, client_id, packet):
        return self._client.send(GTCP_CMD_NOTIFY + client_id + packet)

    @property
    def id(self):
        return self._id

    def on_init(self):
        '''
        to be overridden
        '''
        return

    def on_packet(self, client, request):
        '''
        to be overridden
        '''
        return ''

    def on_client_connect(self, client):
        '''
        to be overridden
        '''
        return

    def on_client_disconnect(self, client):
        '''
        to be overridden
        '''
        return

    def on_background(self):
        '''
        to be overridden
        '''
        return

class GTcpServer(TCPServer):

    def __init__(self, config, processor_class):
        if config.DEBUG:
            from threading import Thread as Process
        else:
            from multiprocessing import Process
        self._config = config
        self._clients = {}
        self._idle_workers = SimpleQueue()
        self._max_queue_size = getattr(config, 'TCP_MAX_QUEUE_SIZE', None)
        self._running_workers = SimpleQueue()
        self._waiting_tasks = SimpleQueue(self._max_queue_size)
        self._worker_processes = []

        max_buffer_size = getattr(config, 'TCP_MAX_BUFFER_SIZE', None)
        self._worker_server = CallbackTcpServer(on_connect=self._on_worker_connect, max_buffer_size=max_buffer_size)
        self._worker_server.listen(**config.WORK_ENDPOINT)
        self._connection_server = CallbackTcpServer(on_connect=self._on_client_connect, max_buffer_size=max_buffer_size)
        for listen_port in config.LISTEN_ENDPOINTS:
            self._connection_server.listen(**listen_port)
        for i in xrange(0, config.WORKER_COUNT):
            processor = processor_class(i, config)
            p = Process(target=processor.run)
            p.processor = processor
            self._worker_processes.append(p)
            p.start()

    def _on_client_connect(self, stream, address):
        client = GTcpConnection(stream, address, self._config, self._on_client_packet,
                                self._on_client_close, endian="!")
        if client.id in self._clients:
            # log.error('tcp_server_dup_client|id=%s,remote=%s', client.id.encode('hex'), client.remote_address)
            log.error('tcp_server_dup_client|id=%s,remote=%s', client.id, client.remote_address)
        self._clients[client.id] = client
        self._handle_task(client, GTCP_CMD_CONNECT, b'')
        # log.info('tcp_server_client_connect|id=%s,remote=%s', client.id.encode('hex'), client.remote_address)
        log.info('tcp_server_client_connect|id=%s,remote=%s', client.id, client.remote_address)

    def _on_worker_connect(self, stream, address):
        worker = GTcpConnection(stream, address, self._config, self._on_worker_packet,
                                self._on_worker_close, endian="<")
        worker.running_task = None
        # log.info('tcp_server_worker_connect|id=%s,remote=%s', worker.id.encode('hex'), worker.remote_address)
        log.info('tcp_server_worker_connect|id=%s,remote=%s', worker.id, worker.remote_address)
        self._on_worker_idle(worker)

    def _handle_task(self, client, cmd, data=b''):
        task = WorkerTask(client, cmd, data)
        if self._idle_workers.empty():
            self._waiting_tasks.put(task)
        else:
            self._assign_task(self._idle_workers.get(), task)

    def _on_client_packet(self, client, data):
        self._handle_task(client, GTCP_CMD_RELAY, data)

    def _on_client_close(self, client):
        # log.info('tcp_server_client_close|id=%s,remote=%s', client.id.encode('hex'), client.remote_address)
        log.info('tcp_server_client_close|id=%s,remote=%s', byteshex(client.id), client.remote_address)
        if client.id not in self._clients:
            # log.error('tcp_server_close_conn_not_found|id=%s,remote=%s', client.id.encode('hex'), client.remote_address)
            log.error('tcp_server_close_conn_not_found|id=%s,remote=%s', byteshex(client.id), client.remote_address)
            return
        self._handle_task(client, GTCP_CMD_DISCONNECT, b'')
        del self._clients[client.id]

    def _on_worker_packet(self, worker, data):
        client = worker.running_task.client
        packet_size = len(data)
        if packet_size < GTCP_HEADER_SIZE:
            # log.error('tcp_worker_reply_error|client_id=%s,client=%s,reply=%s', client.id.encode('hex'), client.remote_address, data.encode('hex'))
            log.error('tcp_worker_reply_error|client_id=%s,client=%s,reply=%s', byteshex(client.id),
                      client.remote_address, byteshex(data))
            worker.running_task.client.close()
            return
        reply_cmd = data[:1]
        if reply_cmd == GTCP_CMD_RELAY or reply_cmd == GTCP_CMD_NOTIFY:
            reply_client = data[GTCP_CMD_SIZE:GTCP_HEADER_SIZE]
            reply_data = data[GTCP_HEADER_SIZE:]
            if reply_client in self._clients:
                self._clients[reply_client].send_packet(reply_data)
            else:
                # log.error('tcp_reply_client_not_found|client_id=%s,reply=%s', reply_client.encode('hex'), reply_data.encode('hex'))
                log.error('tcp_reply_client_not_found|client_id=%s,reply=%s', byteshex(reply_client),
                          byteshex(reply_data))
                worker.running_task.client.close()
        if reply_cmd != GTCP_CMD_NOTIFY:
            self._on_worker_idle(worker)

    def _on_worker_close(self, worker):
        if worker.running_task is not None:
            worker.running_task.client.close()
        self._idle_workers.remove(worker)
        self._running_workers.remove(worker)

    def _on_worker_idle(self, worker):
        if self._waiting_tasks.empty():
            self._idle_workers.put(worker)
        else:
            task = self._waiting_tasks.get()
            self._assign_task(worker, task)

    def _assign_task(self, worker, task):
        worker.running_task = task
        worker.send_packet(task.cmd + task.client.id + task.packet)
        self._running_workers.put(worker)
        #TODO: set timeout

def run():
    IOLoop.instance().start()



from google.protobuf import message
from google.protobuf.descriptor import FieldDescriptor
from sakura import jsonutils
import re
import sys
from six import iteritems
_PY2 = sys.version_info[0] == 2

if not _PY2:
    long = int
    unicode = str
    byteshex = lambda b: b.encode("hex")
    bytesunhex = lambda b: b.decode("hex")
else:
    import binascii
    byteshex = lambda b: binascii.hexlify(b)
    bytesunhex = lambda b: binascii.unhexlify(b)

TYPE_CALLABLE_MAP = {
    FieldDescriptor.TYPE_DOUBLE: float,
    FieldDescriptor.TYPE_FLOAT: float,
    FieldDescriptor.TYPE_INT32: int,
    FieldDescriptor.TYPE_INT64: long,
    FieldDescriptor.TYPE_UINT32: int,
    FieldDescriptor.TYPE_UINT64: long,
    FieldDescriptor.TYPE_SINT32: int,
    FieldDescriptor.TYPE_SINT64: long,
    FieldDescriptor.TYPE_FIXED32: int,
    FieldDescriptor.TYPE_FIXED64: long,
    FieldDescriptor.TYPE_SFIXED32: int,
    FieldDescriptor.TYPE_SFIXED64: long,
    FieldDescriptor.TYPE_BOOL: bool,
    FieldDescriptor.TYPE_STRING: unicode,
    FieldDescriptor.TYPE_BYTES: byteshex,
    FieldDescriptor.TYPE_ENUM: int,
}

def repeated(type_callable):
    return lambda value_list: [type_callable(value) for value in value_list]

def enum_label_name(field, value):
    return field.enum_type.values_by_number[int(value)].name

def pb_to_dict(pb, type_callable_map=TYPE_CALLABLE_MAP, use_enum_labels=False):
    result_dict = {}
    for field, value in pb.ListFields():
        if field.type == FieldDescriptor.TYPE_MESSAGE:
            # recursively encode protobuf sub-message
            type_callable = lambda pb: pb_to_dict(pb,
                type_callable_map=type_callable_map,
                use_enum_labels=use_enum_labels)
        elif field.type in type_callable_map:
            type_callable = type_callable_map[field.type]
        else:
            type_callable = lambda value: '!unknow_type:' + repr(value)
        if use_enum_labels and field.type == FieldDescriptor.TYPE_ENUM:
            type_callable = lambda value: enum_label_name(field, value)
        if field.label == FieldDescriptor.LABEL_REPEATED:
            type_callable = repeated(type_callable)
        result_dict[field.name] = type_callable(value)
    return result_dict

def pb_to_str(pb):
    return jsonutils.to_json(pb_to_dict(pb))

def is_pb_message(data):
    return isinstance(data, message.Message)

def list_all_fields(pb):
    return (i.name for i in pb.DESCRIPTOR.fields)

def set_pb_from_dict(pb, dict_instance, exclude=None):
    fields = set(dict_instance.keys())
    if exclude:
        fields = fields.difference(exclude)
    for field in fields:
        setattr(pb, field, dict_instance[field])
    return pb

def set_pb_from_model(pb, model_instance, exclude=None):
    model_fields = model_instance._meta.get_all_field_names()
    common_fields = set(model_fields).intersection(set(list_all_fields(pb)))
    if exclude:
        common_fields = common_fields.difference(exclude)
    for i in common_fields:
        setattr(pb, i, getattr(model_instance, i))
    return pb

def set_repeated_from_model(pb, queryset, exclude=None):
    if not queryset:
        return pb
    model_fields = queryset[0]._meta.get_all_field_names()
    pb_fields = (i.name for i in pb._message_descriptor.fields)
    common_fields = set(model_fields).intersection(set(pb_fields))
    if exclude:
        common_fields = common_fields.difference(exclude)
    for one_model in queryset:
        one_pb = pb.add()
        for field in common_fields:
            setattr(one_pb, field, getattr(one_model, field))
    return pb

def pb_enum_values_to_class(values, name):
    fields = dict((v.name, v.number) for v in values)
    fields['NAME_TO_VALUE'] = fields.copy()
    fields['VALUE_TO_NAME'] = dict((v.number, v.name) for v in values)
    fields['MIN_VALUE'] = min(v.number for v in values)
    fields['MAX_VALUE'] = max(v.number for v in values)
    return type(name, (object,), fields)

def pb_enum_to_class(pb, name):
    return pb_enum_values_to_class(pb.DESCRIPTOR.enum_types_by_name[name].values, name)


# Protobuf Validator

class _ItemValidator(object):

    def validate(self, path, value):
        if self._validate(value):
            return None
        return self._format_error(path, value)

    def _format_error(self, path, value):
        return self._error_template % (path, value)

class _RequiredValidator(_ItemValidator):

    def __init__(self, field_name):
        self._field_name = field_name
        self._error_template = 'required:%s,%s,%s'

    def _validate(self, value):
        return value.HasField(self._field_name)

    def _format_error(self, path, value):
        return self._error_template % (self._field_name, path, '')

class _MaxItemValidator(_ItemValidator):

    def __init__(self, ref):
        self._ref = ref
        self._error_template = 'maxItem:%s,%%s,%%s' % ref

    def _validate(self, value):
        return len(value) <= self._ref

    def _format_error(self, path, value):
        return self._error_template % (path, len(value))

class _MinItemValidator(_ItemValidator):

    def __init__(self, ref):
        self._ref = ref
        self._error_template = 'minItem:%s,%%s,%%s' % ref

    def _validate(self, value):
        return len(value) >= self._ref

    def _format_error(self, path, value):
        return self._error_template % (path, len(value))

class _MaximumValidator(_ItemValidator):

    def __init__(self, ref):
        self._ref = ref
        self._error_template = 'maximum:%s,%%s,%%s' % ref

    def _validate(self, value):
        return value <= self._ref

class _MinimumValidator(_ItemValidator):

    def __init__(self, ref):
        self._ref = ref
        self._error_template = 'minimum:%s,%%s,%%s' % ref

    def _validate(self, value):
        return value >= self._ref

class _LengthValidator(_ItemValidator):

    def __init__(self, ref):
        self._ref = ref
        self._error_template = 'length:%s,%%s,%%s' % ref

    def _validate(self, value):
        return len(value) == self._ref

class _MaxLengthValidator(_ItemValidator):

    def __init__(self, ref):
        self._ref = ref
        self._error_template = 'maxLength:%s,%%s,%%s' % ref

    def _validate(self, value):
        return len(value) <= self._ref

class _MinLengthValidator(_ItemValidator):

    def __init__(self, ref):
        self._ref = ref
        self._error_template = 'minLength:%s,%%s,%%s' % ref

    def _validate(self, value):
        return len(value) >= self._ref

class _PatternValidator(_ItemValidator):

    def __init__(self, pattern):
        if not pattern.endswith('$'):
            pattern += '$'
        self._pattern = re.compile(pattern)
        self._error_template = 'pattern:%s,%s,%s'

    def _validate(self, value):
        return self._pattern.match(value) is not None

    def _format_error(self, path, value):
        return self._error_template % (self._pattern.pattern, path, value)

def _create_validator(schema):
    validators = {'v': []}
    for key, value in iteritems(schema):
        if key == 'type':
            validators['t'] = value
        elif key == 'properties':
            validators['f'] = {}
            for pname, pvalue in iteritems(value):
                validators['f'][pname] = _create_validator(pvalue)
        elif key == 'required':
            for field_name in value:
                if schema.get('properties', {}).get(field_name, {}).get('type') != 'array':
                    validators['v'].append(_RequiredValidator(field_name))
        elif key == 'items':
            validators['i'] = _create_validator(value)
        elif key == 'maxItems':
            validators['v'].append(_MaxItemValidator(value))
        elif key == 'minItems':
            validators['v'].append(_MinItemValidator(value))
        elif  key == 'maximum':
            validators['v'].append(_MaximumValidator(value))
        elif  key == 'minimum':
            validators['v'].append(_MinimumValidator(value))
        elif  key == 'length':
            validators['v'].append(_LengthValidator(value))
        elif key == 'maxLength':
            validators['v'].append(_MaxLengthValidator(value))
        elif key == 'minLength':
            validators['v'].append(_MinLengthValidator(value))
        elif key == 'pattern':
            validators['v'].append(_PatternValidator(value))
    return validators

def _validate_iter_errors(instance, validator, path='.'):
    try:
        for v in validator['v']:
            result = v.validate(path, instance)
            if result is not None:
                yield result
        if 'f' in validator:
            for fname, fvalidator in iteritems(validator['f']):
                if fvalidator.get('t') == 'array' or instance.HasField(fname):
                    fvalue = getattr(instance, fname)
                    for error in _validate_iter_errors(fvalue, fvalidator, path + '/' + fname):
                        yield error
        if 'i' in validator:
            ivalidator = validator['i']
            for index, fvalue in enumerate(instance):
                for error in _validate_iter_errors(fvalue, ivalidator, path + '/' + str(index)):
                    yield error
    except Exception as ex:
        yield 'exception:' + str(ex)

class PBValidator(object):
    """
    supported schema
    common: type
    object: properties, required
    array: items, maxItems, minItems
    integer: minimum, maximum
    number: minimum, maximum
    string: length, minLength, maxLength, pattern
    bytes: length, minLength, maxLength
    if array field included in required validation, please indicate its type is array
    """

    def __init__(self, schema):
        self.schema = schema

    @property
    def schema(self):
        return self._schema

    @schema.setter
    def schema(self, schema):
        self._schema = schema
        self._validator = _create_validator(self._schema)

    def iter_errors(self, instance):
        for error in _validate_iter_errors(instance, self._validator):
            yield error

    def validate(self, instance):
        for error in self.iter_errors(instance):
            raise error

    def is_valid(self, instance):
        error = next(self.iter_errors(instance), None)
        return error is None



import tornado
from tornado.ioloop import IOLoop
from functools import partial

def create():
    IOLoop().make_current()

def run():
    IOLoop.current().start()

def add_callback(func, *args, **kwargs):
    IOLoop.current().add_callback(func, *args, **kwargs)

def add_timeout(seconds, func, *args, **kwargs):
    return IOLoop.current().call_later(seconds, func, *args, **kwargs)

def remove_timeout(timeout):
    IOLoop.current().remove_timeout(timeout)

def add_interval_timer(seconds, func, *args, **kwargs):
    timer = tornado.ioloop.PeriodicCallback(partial(func, *args, **kwargs), seconds * 1000, IOLoop.instance())
    timer.start()
    return timer



import platform

if platform.system().lower() == 'windows':

    import win32api
    import win32event

    class Lock:

        def __init__(self, name):
            self.name = name
            self.handle = win32event.CreateMutex(None, 0, self.name)
            self.locked = False

        def __del__(self):
            if self.locked:
                self.release()
            win32api.CloseHandle(self.handle)

        def acquire(self):
            win32event.WaitForSingleObject(self.handle, win32event.INFINITE)
            self.locked = True

        def release(self):
            win32event.ReleaseMutex(self.handle)
            self.locked = False

else:

    import os
    import fcntl

    class Lock:

        def __init__(self, name):
            self.filename = os.path.join('/tmp/', name)
            self.handle = open(self.filename, 'a+')
            self.locked = False

        def __del__(self):
            if self.locked:
                self.release()
            self.handle.close()

        # Bitwise OR fcntl.LOCK_NB if you need a non-blocking lock
        def acquire(self):
            fcntl.flock(self.handle.fileno(), fcntl.LOCK_EX)
            self.locked = True

        def release(self):
            fcntl.flock(self.handle.fileno(), fcntl.LOCK_UN)
            self.locked = False



from setuptools import setup

setup(
    name='sakura',
    packages=['sakura'],
    version='0.1.8',
    description='sakura',
    author='None',
    author_email='None',
    maintainer='sakura',
    maintainer_email='sakura@nonono.rocks',
    url='https://github.com/nevermoreluo/sakura',
    download_url='https://github.com/nevermoreluo/sakura/archive/master.zip',
    install_requires=[
        'pycrypto>=2.6.1',
        'python-dateutil>=2.6.1',
        'pytz==2017.2',
        'psutil>=5.3.1',
        'protobuf>=3.4.0',
        'tornado>=4.5.2'
      ],
)



import struct

class BufferReader():

    def __init__(self,  buf, endian='<'):
        self._buffer = buf
        self._length = len(buf)
        self._endian = endian
        self._offset = 0
        self._error = False

    @property
    def error(self):
        return self._error

    def set_endian(self, endian):
        self._endian = endian

    def get_int8(self):
        return self._get_int('b', 1)

    def get_uint8(self):
        return self._get_int('B', 1)

    def get_int16(self):
        return self._get_int('h', 2)

    def get_uint16(self):
        return self._get_int('H', 2)

    def get_int32(self):
        return self._get_int('i', 4)

    def get_uint32(self):
        return self._get_int('I', 4)

    def get_int64(self):
        return self._get_int('q', 8)

    def get_uint64(self):
        return self._get_int('Q', 8)

    def get_float(self):
        return self._get_int('f', 4)

    def get_double(self):
        return self._get_int('d', 8)

    def get_buffer(self, size):
        end_offset = self._offset + size
        if end_offset > self._length:
            self._error = True
        v = self._buffer[self._offset:end_offset]
        self._offset = end_offset
        return v

    def get_remain(self):
        v = self._buffer[self._offset:]
        self._offset += len(v)
        return v

    def skip(self, size):
        self._offset += size
        if self._offset > self._length:
            self._error = True

    def _get_int(self, fmt, size):
        if self._offset + size > self._length:
            self._offset += size
            self._error = True
            return 0
        v = struct.unpack_from(self._endian + fmt, self._buffer, self._offset)[0]
        self._offset += size
        return v


from collections import deque

class SimpleQueue():

    def __init__(self, max_size=None):
        self._q = deque(maxlen=max_size)

    def size(self):
        return len(self._q)

    def empty(self):
        return len(self._q) <= 0

    def put(self, item):
        self._q.append(item)

    def get(self):
        return self._q.popleft()

    def remove(self, value):
        if value in self._q:
            self._q.remove(value)




def pack_archive(dir_name, filter_re=None, filter_dir_re=None):
    import zipfile
    import os
    import re

    def filter_dir(dirpath, filter_dir_re_obj):
        dir_list = dirpath.split(os.path.sep)
        for idx, dirname in enumerate(dir_list):
            if filter_dir_re_obj.match(dirname) is not None:
                if idx == len(dir_list) - 1:
                    return False
                else:
                    return None
        return True

    def filter_file(filename, filter_re_obj):
        return filter_re_obj.match(filename) is None

    if filter_re is None:
        filter_re = '.*\.zip$|.*\.pyc$|.*\.log.*|(^\.)(?!gitkeep)'
    filter_re_obj = re.compile(filter_re)

    if filter_dir_re is None:
        filter_dir_re = '\..+'
    filter_dir_re_obj = re.compile(filter_dir_re)

    archive_name = dir_name + '.zip'
    print('Packing', archive_name)
    f = zipfile.ZipFile(archive_name, 'w', zipfile.ZIP_DEFLATED)
    for dirpath, dirnames, filenames in os.walk('.'):
        is_filter = filter_dir(dirpath, filter_dir_re_obj)
        if is_filter:
            for filename in filenames:
                fullpath = os.path.join(dirpath,filename)
                if filter_file(filename, filter_re_obj):
                    print('Include:', fullpath)
                    f.write(fullpath, os.path.join(dir_name, fullpath))
                else:
                    print('Exclude:', fullpath)
        elif is_filter is not None:
            print('Exclude:', dirpath)
    f.close()
    print('Pack Complete.')



#!/usr/bin/env python

import sys, os, atexit
import signal
import platform
from sakura import osutils


class Daemon:
    """
    A generic daemon class.

    Usage: subclass the Daemon class and override the run() method
    """
    def __init__(self, handler, pidfile, output='/dev/null'):
        self.handler = handler
        self.pidfile = pidfile
        self.output = output

    def daemonize(self):
        """
        do the UNIX double-fork magic, see Stevens' "Advanced
        Programming in the UNIX Environment" for details (ISBN 0201563177)
        http://www.erlenstar.demon.co.uk/unix/faq_2.html#SEC16
        """
        #judge platform
        if platform.system().lower() == 'windows':
            atexit.register(self.delpid)
            pid = str(os.getpid())
            open(self.pidfile,'w+').write("%s\n" % pid)
            return

        #do first fork
        try:
            pid = os.fork()
            if pid > 0:
                # exit first parent
                sys.exit(0)
        except OSError as e:
            sys.stderr.write("fork #1 failed: %d (%s)\n" % (e.errno, e.strerror))
            sys.exit(1)

        # decouple from parent environment
        #os.chdir("/")
        os.setsid()
        os.umask(0)

        #ignore signal
        try:
            signal.signal(signal.SIGINT, signal.SIG_IGN)
            signal.signal(signal.SIGHUP, signal.SIG_IGN)
            signal.signal(signal.SIGQUIT, signal.SIG_IGN)
            signal.signal(signal.SIGPIPE, signal.SIG_IGN)
            signal.signal(signal.SIGTTOU, signal.SIG_IGN)
            signal.signal(signal.SIGTTIN, signal.SIG_IGN)
            #signal.signal(signal.SIGCHLD, signal.SIG_IGN)    #Can't ignore this signal, otherwise popen may be abnormal
        except Exception as ex:
            sys.stderr.write("ignore signal fail: %s\n" % ex)

        # do second fork
        try:
            pid = os.fork()
            if pid > 0:
                # exit from second parent
                sys.exit(0)
        except OSError as e:
            sys.stderr.write("fork #2 failed: %d (%s)\n" % (e.errno, e.strerror))
            sys.exit(1)

        # redirect standard file descriptors
        output_dir = os.path.dirname(self.output)
        if output_dir and not os.path.exists(output_dir):
            os.makedirs(output_dir)
        sys.stdout.flush()
        sys.stderr.flush()
        si = open('/dev/null', 'r')
        so = open(self.output, 'a+')
        os.dup2(si.fileno(), sys.stdin.fileno())
        os.dup2(so.fileno(), sys.stdout.fileno())
        os.dup2(so.fileno(), sys.stderr.fileno())

        # write pidfile
        atexit.register(self.delpid)
        pid = str(os.getpid())
        open(self.pidfile,'w+').write("%s\n" % pid)

    def delpid(self):
        os.remove(self.pidfile)

    def start(self):
        """
        Start the daemon
        """
        # Check for a pidfile to see if the daemon already runs
        try:
            pf = open(self.pidfile,'r')
            pid = int(pf.read().strip())
            pf.close()
        except IOError:
            pid = None

        if pid:
            message = "pidfile %s already exist. Daemon already running?\n"
            sys.stderr.write(message % self.pidfile)
            sys.exit(1)

        # Start the daemon
        self.daemonize()
        self.run()

    def stop(self):
        """
        Stop the daemon
        """
        # Get the pid from the pidfile
        try:
            pf = open(self.pidfile,'r')
            pid = int(pf.read().strip())
            pf.close()
        except IOError:
            pid = None

        if not pid:
            message = "pidfile %s does not exist. Daemon not running?\n"
            sys.stderr.write(message % self.pidfile)
            return # not an error in a restart

        # Try killing the daemon process
        try:
            osutils.kill_process_tree(pid)
            if os.path.exists(self.pidfile):
                os.remove(self.pidfile)
        except Exception as err:
            print(str(err))
            sys.exit(1)

    def restart(self):
        """
        Restart the daemon
        """
        self.stop()
        self.start()

    def run(self):
        """
        Call handler. It will be called after the process has been
        daemonized by start() or restart().
        """
        if self.handler is not None:
            self.handler()

    def main(self):
        if len(sys.argv) == 2:
            action = sys.argv[1].lower()
            if 'start' == action:
                self.start()
            elif 'stop' == action:
                self.stop()
            elif 'restart' == action:
                self.restart()
            else:
                print("Unknown Command")
                print("Usage: %s start|stop|restart" % sys.argv[0])
                sys.exit(2)
            sys.exit(0)
        else:
            print("Usage: %s start|stop|restart" % sys.argv[0])
            sys.exit(2)
